import numpy as np
import pytest

from svgdatagrabber.parser import SvgPathParser


@pytest.mark.parametrize(
    "xrange, yrange, size",
    [
        ((-np.inf, np.inf), (-np.inf, np.inf), 19348),
        ((0, 3000), (5000, 6100), 4608),
        ((3300, 5200), (4800, 6100), 4580),
        ((0, 3000), (3300, 4500), 4488),
        ((3300, 5200), (3300, 4500), 4568),
    ],
)
def test_praai2013(xrange, yrange, size):
    paths = SvgPathParser(
        "praai2013.svg",
        xrange=xrange,
        yrange=yrange,
        min_segments=6,
        drop_horizontal_lines=False,
        drop_vertical_lines=False,
        tolerance=1,
    ).parse()
    assert np.sum([array.size for array in paths.arrays]) == size



from __future__ import annotations

from typing import Iterator

import numpy as np
from qtpy.QtCore import QLineF
from svgpathtools import Line as SvgPathToolsLine

from .exceptions import NotDrawableGeometryError
from .linebase import StraightLineBase
from .point import Point, PointType, Vector


class Line(StraightLineBase):
    """A class representing a line."""

    #: Coefficient of the x term.
    A: float
    #: Coefficient of the y term.
    B: float
    #: Constant term.
    C: float

    def __init__(
        self,
        start: PointType = None,
        end: PointType = None,
        A: float = None,
        B: float = None,
        C: float = None,
        slope: float = None,
        angle: float = None,
        intercept: float = None,
    ):
        """Create a line. Possible ways to create a line (in order of precedence):

        - start and end points (start and end)
        - coefficients A, B, C (A, B and C)
        - start point and slope (start and slope)
        - start point and angle (start and angle)
        - slope and intercept (slope and intercept)
        - angle and intercept (angle and intercept)

        >>> Line(start=Point(0.0, 0.0), end=Point(1.0, 1.0))
        Line(A=1.0, B=-1.0, C=0.0)

        >>> Line(A=1.0, B=-1.0, C=0.0)
        Line(A=1.0, B=-1.0, C=0.0)

        >>> Line(start=Point(0.0, 0.0), slope=1.0)
        Line(A=1.0, B=-1.0, C=0.0)

        >>> Line(start=Point(0.0, 0.0), angle=np.pi / 4.0)
        Line(A=1.0, B=-1.0, C=0.0)

        >>> Line(slope=1.0, intercept=0.0)
        Line(A=1.0, B=-1.0, C=0.0)

        >>> Line(angle=np.pi / 4.0, intercept=0.0)
        Line(A=1.0, B=-1.0, C=0.0)

        >>> Line(start=Point(0.0, 0.0))
        Traceback (most recent call last):
        ...
        ValueError: Not enough information to create a line.

        Args:
            start: The start point of the line.
            end: The end point of the line.
            A: The coefficient of the x term.
            B: The coefficient of the y term.
            C: The constant term.
            slope: The slope of the line.
            intercept: The y-intercept of the line.
        """
        if start is not None and end is not None:
            A, B, C = self.coefficientsFromTwoPoints(start, end)
        elif A is not None and B is not None and C is not None:
            A, B, C = self.standardizeCoefficients(A, B, C)
        elif start is not None and slope is not None:
            A, B, C = self.coefficientsFromPointAndSlope(start, slope)
        elif start is not None and angle is not None:
            A, B, C = self.coefficientsFromPointAndAngle(start, angle)
        elif slope is not None and intercept is not None:
            A, B, C = self.coefficientsFromSlopeAndIntercept(slope, intercept)
        elif angle is not None and intercept is not None:
            A, B, C = self.coefficientsFromAngleAndIntercept(angle, intercept)
        else:
            raise ValueError("Not enough information to create a line.")
        self.A, self.B, self.C = A, B, C

    def __repr__(self) -> str:
        """Return the representation of the line.

        >>> Line(A=1.0, B=-1.0, C=0.0)
        Line(A=1.0, B=-1.0, C=0.0)
        """
        A, B, C = round(self.A, 10), round(self.B, 10), round(self.C, 10)
        return f"{self.__class__.__name__}(A={A}, B={B}, C={C})"

    def __eq__(self, other: Line) -> bool:
        """Return whether the line is equal to another line.

        >>> Line(A=1.0, B=1.0, C=0.0) == Line(A=-1.0, B=-1.0, C=0.0)
        True
        >>> Line(A=1.0, B=1.0, C=0.0) == Line(A=1.0, B=2.0, C=1.0)
        False
        """
        return np.allclose([self.A, self.B, self.C], [other.A, other.B, other.C], atol=self.tolerance)

    def __contains__(self, p: PointType) -> bool:
        """Check if a point is on this line.

        >>> Point(0.0, 0.0) in Line(A=-1.0, B=1.0, C=0.0)
        True
        >>> Point(0.0, 0.0) not in Line(A=-1.0, B=1.0, C=0.0)
        False
        >>> Point(1.0, 0.0) in Line(A=-1.0, B=1.0, C=0.0)
        False
        >>> Point(1.0, 0.0) not in Line(A=-1.0, B=1.0, C=0.0)
        True

        Returns:
            True if the point is on this line, otherwise False.
        """
        p = Point.aspoint(p)
        return self.distance(p) < self.tolerance

    @classmethod
    def standardizeCoefficients(cls, A: float, B: float, C: float) -> tuple[float, float, float]:
        """Standardize the coefficients of a line.

        >>> Line.standardizeCoefficients(1.0, -1.0, 0.0)
        (1.0, -1.0, 0.0)
        >>> Line.standardizeCoefficients(-1.0, -1.0, 1.0)
        (1.0, 1.0, -1.0)

        Args:
            A: The coefficient of the x term.
            B: The coefficient of the y term.
            C: The constant term.

        Returns:
            The standardized coefficients.
        """
        if A != 0.0:
            A, B, C = 1.0, B / A, C / A
        elif B != 0.0:
            B, C = 1.0, C / B
        A, B, C = round(A + 0.0, 10), round(B + 0.0, 10), round(C + 0.0, 10)  # Prevent -0.0 and convert to float
        return A, B, C

    @classmethod
    def fromCoefficients(cls, A: float, B: float, C: float) -> Line:
        """Create a line from the coefficients.

        >>> Line.fromCoefficients(1.0, -1.0, 0.0)
        Line(A=1.0, B=-1.0, C=0.0)

        Args:
            A: The coefficient of the x term.
            B: The coefficient of the y term.
            C: The constant term.

        Returns:
            The created line.
        """
        A, B, C = cls.standardizeCoefficients(A, B, C)
        return cls(A=A, B=B, C=C)

    @classmethod
    def coefficientsFromTwoPoints(cls, p1: PointType, p2: PointType) -> tuple[float, float, float]:
        """Get the coefficients of a line from two points.

        >>> Line.coefficientsFromTwoPoints(Point(0.0, 0.0), Point(1.0, 1.0))
        (1.0, -1.0, 0.0)

        Args:
            p1: The first point.
            p2: The second point.

        Returns:
            The coefficients of the line.
        """
        p1, p2 = Point.aspoint(p1), Point.aspoint(p2)
        A = p1.y - p2.y
        B = p2.x - p1.x
        C = p1.x * p2.y - p2.x * p1.y
        A, B, C = cls.standardizeCoefficients(A, B, C)
        return A, B, C

    @classmethod
    def fromTwoPoints(cls, start: PointType, end: PointType) -> Line:
        """Create a line from two points.

        >>> Line.fromTwoPoints(Point(0.0, 0.0), Point(1.0, 1.0))
        Line(A=1.0, B=-1.0, C=0.0)

        Args:
            start: The first point.
            end: The second point.

        Returns:
            The created line.
        """
        start, end = Point.aspoint(start), Point.aspoint(end)
        A, B, C = cls.coefficientsFromTwoPoints(start, end)
        return cls.fromCoefficients(A=A, B=B, C=C)

    @classmethod
    def coefficientsFromPointAndSlope(cls, p: PointType, slope: float) -> tuple[float, float, float]:
        """Get the coefficients of a line from a point and a slope.

        >>> Line.coefficientsFromPointAndSlope(Point(0.0, 0.0), 1.0)
        (1.0, -1.0, 0.0)

        Args:
            p: The point.
            slope: The slope.

        Returns:
            The coefficients of the line.
        """
        p = Point.aspoint(p)
        A = slope
        B = -1.0
        C = p.y - slope * p.x
        A, B, C = cls.standardizeCoefficients(A, B, C)
        return A, B, C

    @classmethod
    def fromPointAndSlope(cls, start: PointType, slope: float) -> Line:
        """Create a line from a point and a slope.

        >>> Line.fromPointAndSlope(Point(0.0, 0.0), 1.0)
        Line(A=1.0, B=-1.0, C=0.0)

        Args:
            start: The point.
            slope: The slope.

        Returns:
            The created line.
        """
        start = Point.aspoint(start)
        A, B, C = cls.coefficientsFromPointAndSlope(start, slope)
        return cls.fromCoefficients(A=A, B=B, C=C)

    @classmethod
    def coefficientsFromPointAndAngle(cls, p: PointType, angle: float) -> tuple[float, float, float]:
        """Get the coefficients of a line from a point and an angle.

        >>> assert Line.coefficientsFromPointAndAngle(Point(0.0, 0.0), np.pi / 4.0) == (1.0, -1.0, 0.0)

        Args:
            p: The point.
            angle: The angle.

        Returns:
            The coefficients of the line.
        """
        p = Point.aspoint(p)
        A = np.cos(angle)
        B = -np.sin(angle)
        C = B * p.y - A * p.x
        A, B, C = cls.standardizeCoefficients(A, B, C)
        return A, B, C

    @classmethod
    def fromPointAndAngle(cls, start: PointType, angle: float) -> Line:
        """Create a line from a point and an angle.

        >>> Line.fromPointAndAngle(Point(0.0, 0.0), np.pi / 4.0)
        Line(A=1.0, B=-1.0, C=0.0)

        Args:
            start: The point.
            angle: The angle.

        Returns:
            The created line.
        """
        start = Point.aspoint(start)
        A, B, C = cls.coefficientsFromPointAndAngle(start, angle)
        return cls.fromCoefficients(A=A, B=B, C=C)

    @classmethod
    def coefficientsFromSlopeAndIntercept(cls, slope: float, intercept: float) -> tuple[float, float, float]:
        """Get the coefficients of a line from a slope and an intercept.

        >>> Line.coefficientsFromSlopeAndIntercept(1.0, 0.0)
        (1.0, -1.0, 0.0)

        Args:
            slope: The slope.
            intercept: The intercept.

        Returns:
            The coefficients of the line.
        """
        A = slope
        B = -1.0
        C = intercept
        A, B, C = cls.standardizeCoefficients(A, B, C)
        return A, B, C

    @classmethod
    def fromSlopeAndIntercept(cls, slope: float, intercept: float) -> Line:
        """Create a line from a slope and intercept.

        >>> Line.fromSlopeAndIntercept(1.0, 0.0)
        Line(A=1.0, B=-1.0, C=0.0)

        Args:
            slope: The slope.
            intercept: The intercept.

        Returns:
            The created line.
        """
        A, B, C = cls.coefficientsFromSlopeAndIntercept(slope, intercept)
        return cls.fromCoefficients(A=A, B=B, C=C)

    @classmethod
    def coefficientsFromAngleAndIntercept(cls, angle: float, intercept: float) -> tuple[float, float, float]:
        """Get the coefficients of a line from an angle and an intercept.

        >>> assert Line.coefficientsFromAngleAndIntercept(np.pi / 4.0, 0.0) == (1.0, -1.0, 0.0)

        Args:
            angle: The angle.
            intercept: The intercept.

        Returns:
            The coefficients of the line.
        """
        A = np.cos(angle)
        B = -np.sin(angle)
        C = intercept * np.sin(angle)
        A, B, C = cls.standardizeCoefficients(A, B, C)
        return A, B, C

    @classmethod
    def fromAngleAndIntercept(cls, angle: float, intercept: float) -> Line:
        """Create a line from an angle and intercept.

        >>> assert Line.fromAngleAndIntercept(np.pi / 4.0, 0.0) == Line(A=1.0, B=-1.0, C=0.0)

        Args:
            angle: The angle.
            intercept: The intercept.

        Returns:
            The created line.
        """
        A, B, C = cls.coefficientsFromAngleAndIntercept(angle, intercept)
        return cls.fromCoefficients(A=A, B=B, C=C)

    @property
    def maxsize(self) -> float:
        raise ValueError("Line does not have a maxsize.")

    def distance(self, p: Point) -> float:
        """Get the distance between a point and this line.

        >>> line = Line(A=1.0, B=1.0, C=0.0)
        >>> point = Point(1.0, 1.0)
        >>> assert np.isclose(line.distance(point), np.sqrt(2.0))

        Args:
            p: The point to get the distance to.

        Returns:
            The distance between the point and this line.
        """
        p = Point.aspoint(p)
        return abs(self.A * p.x + self.B * p.y + self.C) / np.sqrt(self.A**2 + self.B**2)

    @property
    def slope(self) -> float:
        """Get the slope of this line.

        >>> assert Line(A=1.0, B=-1.0, C=0.0).slope == 1.0
        >>> assert Line(A=1.0, B=1.0, C=0.0).slope == -1.0
        >>> assert Line(A=1.0, B=0.0, C=0.0).slope == np.inf

        Returns:
            The slope of this line.
        """
        if abs(self.B) < self.tolerance:
            return np.inf if self.A > 0 else -np.inf
        return -self.A / self.B

    @property
    def angle(self) -> float:
        """Get the angle of this line.

        >>> angle = Line(A=1.0, B=1.0, C=0.0).angle
        >>> assert np.isclose(angle, -np.pi / 4.0)

        Returns:
            The angle of this line.
        """
        return np.arctan2(-self.A, self.B) + 0.0

    def angleBetween(self, other: Line) -> float:
        """Get the angle between this line and another line.

        >>> line1 = Line(A=1.0, B=1.0, C=0.0)
        >>> line2 = Line(A=1.0, B=-1.0, C=0.0)
        >>> assert line1.angleBetween(line2) == np.pi / 2.0

        Args:
            other: The other line.

        Returns:
            The angle between the two lines.
        """
        return np.arctan2(self.A * other.B - self.B * other.A, self.A * other.A + self.B * other.B) % np.pi + 0.0

    @property
    def intercept(self) -> float:
        """Get the intercept of this line.

        >>> assert Line(A=1.0, B=1.0, C=1.0).intercept == -1.0
        >>> Line(A=1.0, B=0.0, C=0.0).intercept
        Traceback (most recent call last):
        ...
        ValueError: The line is vertical and has no intercept.

        Returns:
            The intercept of this line.
        """
        if abs(self.B) < self.tolerance:
            raise ValueError("The line is vertical and has no intercept.")
        return self.gety(0.0)

    def getx(self, y: float) -> float:
        """Get the x coordinate of a point on this line.

        >>> assert Line(A=1.0, B=1.0, C=0.0).getx(1.0) == -1.0
        >>> Line(A=0.0, B=1.0, C=0.0).getx(0.0)
        Traceback (most recent call last):
        ...
        ValueError: Line is vertical

        Args:
            y: The y coordinate of the point.

        Returns:
            The x coordinate of the point.
        """
        if abs(self.A) < self.tolerance:
            raise ValueError("Line is vertical")
        return -(self.B * y + self.C) / self.A

    def gety(self, x: float) -> float:
        """Get the y coordinate of a point on this line.

        >>> assert Line(A=1.0, B=1.0, C=0.0).gety(1.0) == -1.0
        >>> Line(A=1.0, B=0.0, C=0.0).gety(0.0)
        Traceback (most recent call last):
        ...
        ValueError: Line is horizontal

        Args:
            x: The x coordinate of the point.

        Returns:
            The y coordinate of the point.
        """
        if abs(self.B) < self.tolerance:
            raise ValueError("Line is horizontal")
        return -(self.A * x + self.C) / self.B

    def isParallel(self, line: "Line") -> bool:
        """Check if this line is parallel to another line.

        >>> Line(A=1.0, B=1.0, C=0.0).isParallel(Line(A=1.0, B=1.0, C=1.0))
        True
        >>> Line(A=1.0, B=1.0, C=0.0).isParallel(Line(A=-1.0, B=1.0, C=1.0))
        False

        Args:
            line: The line to check.

        Returns:
            True if the lines are parallel, otherwise False.
        """
        return abs(self.A * line.B - self.B * line.A) < self.tolerance

    def isIntersecting(self, line: "Line") -> bool:
        """Check if this line is intersecting another line.

        >>> Line(A=1.0, B=1.0, C=0.0).isIntersecting(Line(A=1.0, B=1.0, C=1.0))
        False
        >>> Line(A=1.0, B=1.0, C=0.0).isIntersecting(Line(A=-1.0, B=1.0, C=1.0))
        True

        Args:
            line: The line to check.

        Returns:
            True if the lines are intersecting, otherwise False.
        """
        try:
            self.intersect(line)
            return True
        except AssertionError:
            return False

    def isPerpendicular(self, line: "Line") -> bool:
        """Check if this line is perpendicular to another line.

        >>> Line(A=1.0, B=1.0, C=0.0).isPerpendicular(Line(A=1.0, B=1.0, C=1.0))
        False
        >>> Line(A=1.0, B=1.0, C=0.0).isPerpendicular(Line(A=-1.0, B=1.0, C=1.0))
        True

        Args:
            line: The line to check.

        Returns:
            True if the lines are perpendicular, otherwise False.
        """
        return abs(self.A * line.A + self.B * line.B) < self.tolerance

    def intersect(self, line: "Line") -> Point:
        """Get the intersection point between this line and another line.

        >>> Line(A=1.0, B=1.0, C=-1.0).intersect(Line(A=1.0, B=-1.0, C=1.0))
        Point(x=0.0, y=1.0)
        >>> Line(A=1.0, B=1.0, C=0.0).intersect(Line(A=1.0, B=1.0, C=1.0))
        Traceback (most recent call last):
        ...
        AssertionError: Lines are parallel.

        Args:
            line: The line to intersect with.

        Returns:
            The intersection point if there is one, otherwise None.
        """
        assert not self.isParallel(line), "Lines are parallel."
        A = np.asarray([[self.A, self.B], [line.A, line.B]])
        b = np.asarray([-self.C, -line.C])
        x, y = np.linalg.solve(A, b)
        p = Point(x, y)
        assert getattr(self, "extended", False) or p in self and p in line, "Lines do not intersect."
        return p

    def parallel(self, p: PointType) -> "Line":
        """Get a parallel line to this line.

        >>> assert Line(A=1.0, B=-1.0, C=0.0).parallel(Point(0.0, 1.0)) == Line(A=1.0, B=-1.0, C=1.0)

        Args:
            p: A point on the parallel line.

        Returns:
            A parallel line to this line.
        """
        p = Point.aspoint(p)
        return Line(A=self.A, B=self.B, C=-self.A * p.x - self.B * p.y)

    def perpendicular(self, p: PointType) -> "Line":
        """Get a perpendicular line to this line.

        >>> assert Line(A=1.0, B=-1.0, C=0.0).perpendicular(Point(0.0, 1.0)) == Line(A=1.0, B=1.0, C=-1.0)
        >>> assert Line(A=0.0, B=1.0, C=0.0).perpendicular(Point(0.0, 0.0)) == Line(A=1.0, B=0.0, C=0.0)
        >>> assert Line(A=1.0, B=0.0, C=0.0).perpendicular(Point(0.0, 0.0)) == Line(A=0.0, B=1.0, C=0.0)

        Args:
            p: A point on the perpendicular line.

        Returns:
            A perpendicular line to this line.
        """
        p = Point.aspoint(p)
        return Line(A=self.B, B=-self.A, C=-self.B * p.x + self.A * p.y)

    @property
    def drawArgs(self):
        raise NotDrawableGeometryError


class LineSegment(Line):
    """A class representing a line segment."""

    #: The first point to create the line.
    start: Point
    #: The second point to create the line.
    end: Point
    #: Whether to extend the segment to infinity, used to check if a point is on the extended segment.
    extended: bool

    def __init__(
        self,
        start: PointType,
        end: PointType,
        extended: bool = False,
    ):
        """Create a new line segment.

        >>> LineSegment(start=Point(0.0, 0.0), end=Point(1.0, 1.0))
        LineSegment(start=Point(x=0.0, y=0.0), end=Point(x=1.0, y=1.0)) -> LineSegment(A=1.0, B=-1.0, C=0.0)

        Args:
            start: The first point to create the line.
            end: The second point to create the line.
            extended: Whether to extend the segment to infinity, used to check if a point is on the extended segment.
        """
        super().__init__(start=start, end=end)
        self.start, self.end = Point.aspoint(start), Point.aspoint(end)
        self.extended = extended

    def __repr__(self):
        """Get the string representation of this line segment.

        >>> repr(LineSegment(start=Point(0.0, 0.0), end=Point(1.0, 1.0)))
        'LineSegment(start=Point(x=0.0, y=0.0), end=Point(x=1.0, y=1.0)) -> LineSegment(A=1.0, B=-1.0, C=0.0)'
        """
        return f"{self.__class__.__name__}(start={self.start}, end={self.end}) -> {super().__repr__()}"

    def __eq__(self, other: LineSegment) -> bool:
        """Check if this line segment is equal to another line segment.

        >>> segment1 = LineSegment(start=Point(0.0, 0.0), end=Point(1.0, 1.0))
        >>> segment2 = LineSegment(start=Point(0.0, 0.0), end=Point(1.0, 1.0))
        >>> segment1 == segment2
        True
        >>> segment3 = LineSegment(start=Point(0.0, 0.0), end=Point(2.0, 2.0))
        >>> segment1 == segment3
        False

        Args:
            other: The line segment to check.
        """
        return super().__eq__(other) and (
            (self.start == other.start and self.end == other.end)
            or (self.start == other.end and self.end == other.start)
        )

    def __contains__(self, p: PointType) -> bool:
        """Check if a point is on this segment.

        >>> Point(0.0, 0.0) in LineSegment(start=Point(0.0, 0.0), end=Point(1.0, 1.0))
        True
        >>> Point(1.0, 1.0) in LineSegment(start=Point(0.0, 0.0), end=Point(1.0, 1.0))
        True
        >>> Point(0.5, 0.5) in LineSegment(start=Point(0.0, 0.0), end=Point(1.0, 1.0))
        True
        >>> Point(0.0, 1.0) in LineSegment(start=Point(0.0, 0.0), end=Point(1.0, 1.0))
        False

        Returns:
            True if the point is on this segment, otherwise False.
        """
        p = Point.aspoint(p)
        minx, maxx = sorted([self.start.x, self.end.x])
        miny, maxy = sorted([self.start.y, self.end.y])
        return super().__contains__(p) and (self.extended or (minx <= p.x <= maxx and miny <= p.y <= maxy))

    def __iter__(self) -> Iterator[Point]:
        """Iterate over the points of this segment.

        >>> list(LineSegment(start=Point(0.0, 0.0), end=Point(1.0, 1.0)))
        [Point(x=0.0, y=0.0), Point(x=1.0, y=1.0)]
        """
        yield self.start
        yield self.end

    @classmethod
    def fromSvgPathToolsLine(cls, line: SvgPathToolsLine):
        """Create a new line segment from a SvgPathTools line.

        >>> LineSegment.fromSvgPathToolsLine(SvgPathToolsLine(start=.0j, end=1.0+1.0j))
        LineSegment(start=Point(x=0.0, y=0.0), end=Point(x=1.0, y=1.0)) -> LineSegment(A=1.0, B=-1.0, C=0.0)

        Args:
            line: The SvgPathTools line to create the line segment from.

        Returns:
            A new line segment.
        """
        return cls(start=line.start, end=line.end)

    @property
    def maxsize(self) -> float:
        """Get the maximal size of this segment.

        >>> assert LineSegment(start=Point(0.0, 0.0), end=Point(3.0, 4.0)).maxsize == 5.0
        """
        return self.length

    @property
    def length(self) -> float:
        """Get the length of this segment.

        >>> assert LineSegment(start=Point(0.0, 0.0), end=Point(3.0, 4.0)).length == 5.0

        Returns:
            The length of this segment.
        """
        return self.start.distance(self.end)

    @property
    def direction(self) -> float:
        """Get the direction of this segment.

        >>> assert LineSegment(start=Point(0.0, 0.0), end=Point(1.0, 1.0)).direction - np.pi / 4 == 0.0
        >>> assert LineSegment(start=Point(0.0, 0.0), end=Point(-1.0, 1.0)).direction - 3.0 * np.pi / 4 == 0.0
        >>> assert LineSegment(start=Point(0.0, 0.0), end=Point(-1.0, -1.0)).direction + 3.0 * np.pi / 4.0 == 0.0
        >>> assert LineSegment(start=Point(0.0, 0.0), end=Point(1.0, -1.0)).direction + np.pi / 4.0 == 0.0

        Returns:
            The direction of this segment.
        """
        return self.start.direction(self.end)

    @property
    def midpoint(self) -> Point:
        """Get the midpoint of this segment.

        >>> assert LineSegment(start=Point(0.0, 0.0), end=Point(1.0, 1.0)).midpoint == Point(x=0.5, y=0.5)

        Returns:
            The midpoint of this segment.
        """
        return Point((self.start.x + self.end.x) / 2, (self.start.y + self.end.y) / 2)

    def reverse(self) -> "LineSegment":
        """Reverse the direction of this segment.

        >>> LineSegment(start=Point(0.0, 0.0), end=Point(1.0, 1.0)).reverse()
        LineSegment(start=Point(x=1.0, y=1.0), end=Point(x=0.0, y=0.0)) -> LineSegment(A=1.0, B=-1.0, C=0.0)
        """
        self.start, self.end = self.end, self.start
        return self

    @property
    def drawArgs(self) -> QLineF:
        """Get the Qt representation of this line segment."""
        return QLineF(self.start.x, self.start.y, self.end.x, self.end.y)


class ExtendedLineSegment(LineSegment):
    """A class representing an extended line segment."""

    def __init__(self, start: PointType, end: PointType):
        """Create a new extended line segment.

        >>> segment = ExtendedLineSegment(start=Point(0.0, 0.0), end=Point(1.0, 1.0))
        >>> assert segment.extended

        Args:
            start: The start point of the segment.
            end: The end point of the segment.
        """
        super().__init__(start=start, end=end, extended=True)


class LineRay(Line):
    """A class representing a line ray."""

    #: The first point to create the line.
    start: Point
    #: The second point to create the line.
    end: Point
    #: Whether to extend the ray to infinity, used to check if a point is on the extended ray.
    extended: bool

    def __init__(
        self,
        start: PointType,
        end: PointType,
        extended: bool = False,
    ):
        """Create a ray.

        >>> LineRay(start=Point(0.0, 0.0), end=Point(1.0, 1.0))
        LineRay(start=Point(x=0.0, y=0.0), slope=1.0) -> LineRay(A=1.0, B=-1.0, C=0.0)

        Args:
            start: The first point to create the line.
            end: The second point to create the line.
            extended: Whether to extend the ray to infinity, used to check if a point is on the extended ray.
        """
        super().__init__(start=start, end=end)
        self.start, self.end = Point.aspoint(start), Point.aspoint(end)
        self.direction = self.end - self.start
        self.extended = extended

    def __repr__(self):
        """Get the string representation of this ray.

        >>> LineRay(start=Point(0.0, 0.0), end=Point(1.0, 1.0))
        LineRay(start=Point(x=0.0, y=0.0), slope=1.0) -> LineRay(A=1.0, B=-1.0, C=0.0)
        """
        slope = round(self.slope, 10)
        return f"{self.__class__.__name__}(start={self.start}, slope={slope}) -> {super().__repr__()}"

    def __eq__(self, other: LineRay) -> bool:
        """Check if two rays are equal.

        >>> LineRay(start=Point(0.0, 0.0), end=Point(1.0, 1.0)) == LineRay(start=Point(0.0, 0.0), end=Point(1.0, 1.0))
        True
        >>> LineRay(start=Point(0.0, 0.0), end=Point(1.0, 1.0)) == LineRay(start=Point(0.0, 0.0), end=Point(2.0, 2.0))
        True
        >>> LineRay(start=Point(0.0, 0.0), end=Point(1.0, 1.0)) == LineRay(start=Point(-1.0, -1.0), end=Point(1.0, 1.0))
        False
        """
        return super().__eq__(other) and self.start == other.start

    def __contains__(self, p: PointType):
        """Check if a point is on this ray.

        >>> Point(0.0, 0.0) in LineRay(start=Point(0.0, 0.0), end=Point(1.0, 1.0))
        True
        >>> Point(0.5, 0.5) in LineRay(start=Point(0.0, 0.0), end=Point(1.0, 1.0))
        True
        >>> Point(1.0, 1.0) in LineRay(start=Point(0.0, 0.0), end=Point(1.0, 1.0))
        True
        >>> Point(2.0, 2.0) in LineRay(start=Point(0.0, 0.0), end=Point(1.0, 1.0))
        True
        >>> Point(-1.0, -1.0) in LineRay(start=Point(0.0, 0.0), end=Point(1.0, 1.0))
        False

        Returns:
            True if the point is on this ray, otherwise False.
        """
        return super().__contains__(p) and (self.extended or self.start.vector(p) @ self.slope_vector >= 0)

    @property
    def slope_vector(self) -> Vector:
        """Get the slope vector of this ray.

        >>> LineRay(start=Point(0.0, 0.0), end=Point(1.0, 1.0)).slope_vector
        Vector(x=1.0, y=1.0)

        Returns:
            The slope vector of this ray.
        """
        return Vector.asvector(self.end - self.start)


class ExtendedLineRay(LineRay):
    """A class representing an extended line ray."""

    def __init__(self, start: PointType, end: PointType):
        """Create an extended ray.

        >>> ray = ExtendedLineRay(start=Point(0.0, 0.0), end=Point(1.0, 1.0))
        >>> assert ray.extended

        Args:
            start: The first point to create the ray.
            end: The second point to create the ray.
        """
        super().__init__(start=start, end=end, extended=True)



from __future__ import annotations

from .point import Point, PointType
from .polygon import Polygon


class Triangle(Polygon):
    """A class representing a triangle."""

    def __init__(self, *points: PointType):
        """Initialize a triangle.

        >>> Triangle(Point(0.0, 0.0), Point(1.0, 0.0), Point(0.0, 1.0))
        Triangle(Point(x=0.0, y=0.0), Point(x=1.0, y=0.0), Point(x=0.0, y=1.0))

        Args:
            *points: The points of the triangle.
        """
        if len(points) != 3:
            raise ValueError("A triangle must have exactly three points.")
        super().__init__(*points)

    @property
    def circumcenter(self) -> Point:
        """Return the circumcenter of the triangle.

        >>> Triangle(Point(0.0, 0.0), Point(1.0, 0.0), Point(0.0, 1.0)).circumcenter
        Point(x=0.5, y=0.5)
        """
        a, b, c = self.vertices  # type: Point
        x = a.x * a.x + a.y * a.y
        y = b.x * b.x + b.y * b.y
        z = c.x * c.x + c.y * c.y
        s = a.x * (b.y - c.y) + b.x * (c.y - a.y) + c.x * (a.y - b.y)
        if abs(s) < self.tolerance:
            raise ValueError("The triangle is degenerate.")
        return Point(
            (x * (b.y - c.y) + y * (c.y - a.y) + z * (a.y - b.y)) / (2 * s),
            (x * (c.x - b.x) + y * (a.x - c.x) + z * (b.x - a.x)) / (2 * s),
        )



from qtpy.QtWidgets import QGraphicsScene

from ..geometry import GeometryBase
from .annotations import QBrushType, QGraphicsItemType, QPenType


class GeometricObject:
    #: The geometry of the object.
    geometry: GeometryBase
    #: The graphics item of the object.
    item: QGraphicsItemType

    def __init__(self, geometry: GeometryBase, item: QGraphicsItemType = None):
        """Create a geometric object.

        Args:
            geometry: The geometry of the object.
            item: The graphics item of the object.
        """
        self.geometry = geometry
        self.item = item

    def redraw(
        self,
        scene: QGraphicsScene,
        pen: QPenType = None,
        brush: QBrushType = None,
    ):
        """Draw the object in the scene."""
        self.item = self.geometry.draw(scene, pen, brush, self.item)



"""
The Graphics User Interface (GUI) for the SVG Data Grabber, run with ``python -m svgdatagrabber.graphics`` or
``svgdatagrabber-graphics``.
"""


def main():
    import sys

    from qtpy.QtWidgets import QApplication

    from svgdatagrabber.graphics.mainwindow import SvgDataGrabberMainWindow

    app = QApplication(sys.argv)
    window = SvgDataGrabberMainWindow()
    window.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import Union

from svgpathtools.path import Arc, CubicBezier, Line, Path, QuadraticBezier

from .filterbase import FilterBase


class HorizontalLineFilter(FilterBase):
    """Filter for horizontal lines."""

    def accept(self, path: Path) -> bool:
        def segmentIsHorizontal(segment: Union[Line, QuadraticBezier, CubicBezier, Arc]):
            if not isinstance(segment, Line):
                return False
            return abs(segment.start.imag - segment.end.imag) < self.tolerance

        return not self.enabled or not (all(segmentIsHorizontal(segment) for segment in path))


class VerticalLineFilter(FilterBase):
    """Filter for vertical lines."""

    def accept(self, path: Path) -> bool:
        def segmentIsVertical(segment: Union[Line, QuadraticBezier, CubicBezier, Arc]):
            if not isinstance(segment, Line):
                return False
            return abs(segment.start.real - segment.end.real) < self.tolerance

        return not self.enabled or not (all(segmentIsVertical(segment) for segment in path))



class QtException(Exception):
    """Base class for all Qt exceptions."""

    pass


class NotDrawableGeometryError(QtException):
    """An exception raised when a geometry cannot be drawn on a Qt scene."""

    pass


class NoCorrespondingQtObjectError(QtException):
    """An exception raised when a geometry has no corresponding Qt object."""

    pass



from __future__ import annotations

from abc import ABC

from svgpathtools.path import Path


class FilterBase(ABC):
    """Base class for filters."""

    #: Enabled or not
    enabled: bool = True
    #: Tolerance for determining if a path is a horizontal or vertical line
    tolerance: float = 1e-6

    def __init__(self, *, enabled: bool = True, tolerance: float = 1e-6):
        self.enabled = enabled
        self.tolerance = tolerance

    def accept(self, path: Path) -> bool:
        """Accept or reject a path.

        Args:
            path: Path to check.

        Returns:
            True if the path is accepted, False otherwise.
        """
        raise NotImplementedError

    def __call__(self, path: Path) -> bool:
        return self.accept(path)

    def enable(self):
        """Enable the filter."""
        self.enabled = True

    def disable(self):
        """Disable the filter."""
        self.enabled = False



from __future__ import annotations

from typing import Callable

from svgpathtools.path import Path

from .filterbase import FilterBase


class CustomFilter(FilterBase):
    """Custom filter."""

    #: Custom filter function
    filter_function: Callable[[Path], bool]

    def __init__(self, filter_function: Callable[[Path], bool], *, enabled: bool = True, tolerance: float = 1e-6):
        super().__init__(enabled=enabled, tolerance=tolerance)
        self.filter_function = filter_function

    def accept(self, path: Path) -> bool:
        return not self.enabled or self.filter_function(path)



from __future__ import annotations

import math
from typing import TYPE_CHECKING, Iterable, List, Union

import numpy as np
from svgpathtools import Arc as SvgPathToolsArc
from svgpathtools import CubicBezier as SvgPathToolsCubicBezier
from svgpathtools import Line as SvgPathToolsLine
from svgpathtools import Path as SvgPathToolsPath
from svgpathtools import QuadraticBezier as SvgPathToolsQuadraticBezier
from typing_extensions import Self

from .arc import Arc
from .bezier import Bezier, CubicBezier, QuadraticBezier
from .csys import CoordinateSystem
from .linebase import LineBase, StraightLineBase
from .sequence import GeometrySequence, LineSequence
from .straightline import LineSegment

SvgPathToolsSegmentType = Union[SvgPathToolsLine, SvgPathToolsArc, SvgPathToolsQuadraticBezier, SvgPathToolsCubicBezier]

if TYPE_CHECKING:
    from matplotlib.axes import Axes
    from pandas import DataFrame


class Path(StraightLineBase, LineSequence):
    items: List[LineSegment | Arc | QuadraticBezier | CubicBezier]

    #: Whether to only export the start and end points of the path.
    start_end_only: bool

    def __init__(self, *segments: LineBase, start_end_only: bool = True):
        LineSequence.__init__(self, *segments)
        self.start_end_only = start_end_only

    def __repr__(self):
        return LineSequence.__repr__(self)

    def __eq__(self, other: "Path") -> bool:
        return LineSequence.__eq__(self, other)

    @property
    def segments(self) -> List[LineSegment | Arc | QuadraticBezier | CubicBezier]:
        return self.items

    @classmethod
    def fromSvgPathToolsPath(cls, path: "SvgPathToolsPath", start_end_only: bool = True):
        segments = []
        for segment in path:
            if isinstance(segment, SvgPathToolsLine):
                segments.append(LineSegment.fromSvgPathToolsLine(segment))
            elif isinstance(segment, SvgPathToolsArc):
                segments.append(Arc.fromSvgPathToolsArc(segment))
            elif isinstance(segment, SvgPathToolsQuadraticBezier):
                segments.append(QuadraticBezier.fromSvgPathToolsQuadraticBezier(segment))
            elif isinstance(segment, SvgPathToolsCubicBezier):
                segments.append(CubicBezier.fromSvgPathToolsCubicBezier(segment))
        return cls(*segments, start_end_only=start_end_only)

    @property
    def array(self) -> np.ndarray:
        if self.start_end_only:
            return np.array([[p.x, p.y] for segment in self.segments for p in [segment.start, segment.end]])
        else:
            return np.array([[p.x, p.y] for segment in self.segments for p in segment])

    def transformed(self, csys: CoordinateSystem):
        for segment in self.items:
            segment.start = csys.transform(segment.start)
            segment.end = csys.transform(segment.end)
            if isinstance(segment, Bezier):
                segment.controls = [csys.transform(control) for control in segment.controls]


class PathSequence(GeometrySequence):
    items: List[Path]

    def __init__(self, *paths: Path):
        GeometrySequence.__init__(self, *paths)

    def __getitem__(self, item) -> Self | Path:
        items = self.items[item]
        return self.__class__(*items) if isinstance(items, list) else items

    @property
    def paths(self) -> List[Path]:
        return self.items

    @classmethod
    def fromSvgPathToolsPathSequence(cls, paths: Iterable["SvgPathToolsPath"]):
        return cls(*[Path.fromSvgPathToolsPath(path) for path in paths])

    @property
    def arrays(self) -> List[np.ndarray]:
        return [path.array for path in self.paths]

    def transformed(self, csys: CoordinateSystem):
        """Transform the paths.

        Args:
            csys: Coordinate system to transform the paths to.

        Returns:
            The transformed paths.
        """
        for path in self.items:
            path.transformed(csys)
        return self

    def plot(self, ax: "Axes" = None, fig_kwargs: dict = None, **kwargs) -> "Axes":
        """Plot the paths.

        Args:
            ax: Axes to plot on.
            fig_kwargs: Keyword arguments to pass to plt.figure().
            kwargs: Keyword arguments to pass to plt.plot().
        """
        import matplotlib.pyplot as plt

        if ax is None:
            _, ax = plt.subplots(**(fig_kwargs or {}))
        arrays = self.arrays
        for idx, point in enumerate(arrays):
            kwargs["label"] = f"Line-{idx}"
            ax.plot(point[:, 0], point[:, 1], **kwargs)
        ax.legend(ncol=math.floor(math.sqrt(len(arrays))))
        ax.grid()
        return ax

    def df(self, x: str = "x", y: str = "y") -> "DataFrame":
        """Get the paths as a pandas DataFrame."""
        import pandas as pd

        df = pd.DataFrame(columns=[x, y, "path"])
        for idx, line in enumerate(self.arrays):
            df = pd.concat([df, pd.DataFrame({x: line[:, 0], y: line[:, 1], "path": idx})], ignore_index=True)
        return df

    def to_csv(self, path: str, x: str = "x", y: str = "y"):
        """Save the paths as a csv file."""
        self.df(x, y).to_csv(path, index=False)



from __future__ import annotations

from typing import List, overload

from qtpy.QtCore import Qt
from qtpy.QtGui import QPainter
from qtpy.QtWidgets import QGraphicsScene, QGraphicsView, QOpenGLWidget, QWidget

from ..geometry import GeometryBase
from .annotations import QBrushType, QPenType
from .geometricobject import GeometricObject


class GraphicsView(QGraphicsView):
    """GraphicsView class."""

    #: geometric objects
    geometric_objects: List[GeometricObject]

    @overload
    def __init__(self, parent: QWidget | None = None, *, useOpenGL: bool = False): ...

    @overload
    def __init__(self, scene: QGraphicsScene = None, parent: QWidget | None = None, *, useOpenGL: bool = False): ...

    def __init__(
        self,
        scene: QGraphicsScene | QWidget | None = None,
        parent: QWidget | None = None,
        *,
        useOpenGL: bool = False,
    ):
        """Initialize the class."""
        if isinstance(scene, QWidget):
            scene, parent = None, scene
        super().__init__(parent)
        self.scene = scene or QGraphicsScene()
        self.setScene(self.scene)
        self.setRenderHint(QPainter.HighQualityAntialiasing)

        # geometric objects
        self.geometric_objects = []

        # use OpenGL
        useOpenGL and self.useOpenGL()

    def addPrimitive(self, primitive: GeometricObject | GeometryBase, fitInView: bool = True, scale: float = 1.0):
        """Add a primitive to the scene."""
        if isinstance(primitive, GeometryBase):
            primitive = GeometricObject(primitive)
        primitive.redraw(self.scene)
        self.geometric_objects.append(primitive)
        fitInView and self.fitInView(self.scene.itemsBoundingRect(), Qt.KeepAspectRatio)
        self.scale(scale, scale)

    def redraw(self, pen: QPenType = None, brush: QBrushType = None, fitInView: bool = True, scale: float = 1.0):
        """Draw the geometries in the scene and fit the view."""
        for geometric_object in self.geometric_objects:
            geometric_object.redraw(self.scene, pen, brush)
        fitInView and self.fitInView(self.scene.itemsBoundingRect(), Qt.KeepAspectRatio)
        self.scale(scale, scale)

    def useOpenGL(self):
        """Use OpenGL."""
        self.setViewport(QOpenGLWidget())

    def wheelEvent(self, event):
        """
        Zoom in or out of the view, from https://stackoverflow.com/a/29026916/18728919.
        """
        # Zoom factors
        zoomInFactor = 1.25
        zoomOutFactor = 1 / zoomInFactor

        # Set Anchors
        self.setTransformationAnchor(QGraphicsView.NoAnchor)
        self.setResizeAnchor(QGraphicsView.NoAnchor)

        # Save the scene pos
        oldPos = self.mapToScene(event.pos())

        # Zoom
        zoomFactor = zoomInFactor if event.angleDelta().y() > 0 else zoomOutFactor
        self.scale(zoomFactor, zoomFactor)
        self.centerOn(self.mapToScene(self.viewport().rect().center()))

        # Get the new position
        newPos = self.mapToScene(event.pos())

        # Move scene to old position
        delta = newPos - oldPos
        self.translate(delta.x(), delta.y())



from abc import ABC

from .geometrybase import DrawAsLine, GeometryBase


class LineBase(GeometryBase, ABC):
    """A base class for line-like geometries."""

    drawAs = DrawAsLine

    pass


class StraightLineBase(LineBase, ABC):
    """A base class for straight line-like geometries."""

    pass


class CurveLineBase(LineBase, ABC):
    """A base class for curve line-like geometries."""

    pass



from __future__ import annotations

from typing import Union

import numpy as np
from svgpathtools.path import Arc, CubicBezier, Line, Path, QuadraticBezier

from .filterbase import FilterBase


class RectangleRangeFilter(FilterBase):
    """Filter paths based on their range."""

    #: Range of x values to include.
    xrange: tuple[float, float]
    #: Range of y values to include.
    yrange: tuple[float, float]
    #: Include or exclude the range
    include: bool
    #: Sensitive or not
    sensitive: bool

    def __init__(
        self,
        xrange: tuple[float, float] = (-np.inf, np.inf),
        yrange: tuple[float, float] = (-np.inf, np.inf),
        include: bool = True,
        sensitive: bool = False,
        enabled: bool = True,
        tolerance: float = 1e-6,
    ):
        super().__init__(enabled=enabled, tolerance=tolerance)
        self.xrange = tuple(xrange)
        self.yrange = tuple(yrange)
        self.include = include
        self.sensitive = sensitive

    def accept(self, path: Path) -> bool:
        def pointInRange(p: complex):
            return self.xrange[0] <= p.real <= self.xrange[1] and self.yrange[0] <= p.imag <= self.yrange[1]

        def segmentInRange(seg: Union[Line, QuadraticBezier, CubicBezier, Arc]):
            func = all if self.sensitive else any
            return func(pointInRange(p) for p in seg.bpoints())

        def pathInRange(pth: Path):
            func = all if self.sensitive else any
            return func(segmentInRange(seg) for seg in pth)

        def pathOutRange(pth: Path):
            return not pathInRange(pth)

        return not self.enabled or (self.include and pathInRange(path)) or (not self.include and pathOutRange(path))



from __future__ import annotations

from abc import abstractmethod
from typing import Iterable, List, Sequence, Union, overload

from .geometrybase import GeometryBase
from .point import Point, PointType
from .straightline import Line

SequenceItem = Union[GeometryBase, Point, Line]
SequenceItemType = Union[GeometryBase, PointType, Line]


class GeometrySequence(Sequence[GeometryBase]):
    """A class representing a sequence of geometries."""

    #: Items in the sequence.
    items: List[SequenceItem]

    def __init__(self, *items: SequenceItemType):
        """Create a sequence of items.

        >>> GeometrySequence(Point(1.0, 2.0), Point(3.0, 4.0))
        GeometrySequence(Point(x=1.0, y=2.0), Point(x=3.0, y=4.0))

        Args:
            items: The items to add to the sequence.
        """
        self.items = list(self.asitem(item) for item in items)
        self.check()

    def __repr__(self) -> str:
        """Return a string representation of the sequence.

        >>> repr(GeometrySequence(Point(1.0, 2.0), Point(3.0, 4.0)))
        'GeometrySequence(Point(x=1.0, y=2.0), Point(x=3.0, y=4.0))'
        """
        return f"{self.__class__.__name__}({', '.join(repr(p) for p in self.items)})"

    @overload
    @abstractmethod
    def __getitem__(self, index: int) -> SequenceItem: ...

    @overload
    @abstractmethod
    def __getitem__(self, index: slice) -> Sequence[SequenceItem]: ...

    def __getitem__(self, index: int) -> SequenceItem | Sequence[SequenceItem]:
        """Get a point from the sequence.

        >>> GeometrySequence(Point(1.0, 2.0), Point(3.0, 4.0))[0]
        Point(x=1.0, y=2.0)
        >>> GeometrySequence(Point(1.0, 2.0), Point(3.0, 4.0))[1]
        Point(x=3.0, y=4.0)
        >>> GeometrySequence(Point(1.0, 2.0), Point(3.0, 4.0))[0:1]
        [Point(x=1.0, y=2.0)]

        Args:
            index: The index of the item to get.
        """
        return self.items[index]

    def __setitem__(self, key: int, value: SequenceItemType):
        """Set a point in the sequence.

        >>> ps = GeometrySequence(Point(1.0, 2.0), Point(3.0, 4.0))
        >>> ps[1] = Point(5.0, 6.0)
        >>> ps
        GeometrySequence(Point(x=1.0, y=2.0), Point(x=5.0, y=6.0))
        """
        self.items[key] = self.asitem(value)

    def __len__(self) -> int:
        """Return the length of the sequence.

        >>> len(GeometrySequence(Point(1.0, 2.0), Point(3.0, 4.0)))
        2
        """
        return len(self.items)

    def __iter__(self) -> Iterable[GeometryBase]:
        """Return an iterator over the items.

        >>> list(GeometrySequence(Point(1.0, 2.0), Point(3.0, 4.0)))
        [Point(x=1.0, y=2.0), Point(x=3.0, y=4.0)]
        """
        return iter(self.items)

    def __reversed__(self):
        """Return a reversed iterator over the items.

        >>> list(reversed(GeometrySequence(Point(1.0, 2.0), Point(3.0, 4.0))))
        [Point(x=3.0, y=4.0), Point(x=1.0, y=2.0)]
        """
        return reversed(self.items)

    def __eq__(self, other: "GeometrySequence") -> bool:
        """Check if two polygons are equal.

        >>> ps1 = GeometrySequence(Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0))
        >>> ps2 = GeometrySequence(Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0))
        >>> ps1 == ps2
        True
        """
        return all(v1 == v2 for v1, v2 in zip(self.items, other.items))

    def __contains__(self, item: SequenceItem | Iterable[SequenceItem]) -> bool:
        """Check if an item is in the sequence.

        >>> Point(1.0, 2.0) in GeometrySequence(Point(0.0, 0.0), Point(1.0, 2.0))
        True
        >>> (Point(1.0, 2.0), Point(3.0, 4.0)) in GeometrySequence(Point(0.0, 0.0), Point(1.0, 2.0))
        False
        >>> (Point(1.0, 2.0), Point(3.0, 4.0)) in GeometrySequence(Point(0.0, 0.0), Point(1.0, 2.0), Point(3.0, 4.0))
        True
        """
        if isinstance(item, Iterable) and isinstance(tuple(item)[0], Point):
            return all(p in self.items for p in item)
        return self.asitem(item) in self.items

    def check(self):
        """Check if the sequence is valid."""
        pass

    @classmethod
    def asitem(cls, item: SequenceItemType) -> SequenceItem:
        """Prepare an item for adding to the sequence, if necessary.

        >>> GeometrySequence.asitem(Point(1.0, 2.0))
        Point(x=1.0, y=2.0)
        """
        return item

    def append(self, item: SequenceItemType) -> None:
        """Add a point to the sequence.

        >>> ps = GeometrySequence(Point(1.0, 2.0), Point(3.0, 4.0))
        >>> ps.append(Point(5.0, 6.0))
        >>> ps
        GeometrySequence(Point(x=1.0, y=2.0), Point(x=3.0, y=4.0), Point(x=5.0, y=6.0))
        """
        self.items.append(self.asitem(item))

    def index(self, item: SequenceItemType, start: int = 0, stop: int = None) -> int:
        """Find the index of a point in the sequence.

        >>> ps = GeometrySequence(Point(1.0, 2.0), Point(3.0, 4.0))
        >>> ps.index(Point(3.0, 4.0))
        1

        Args:
            item: The item to find.
            start: The start index.
            stop: The end index.

        Returns:
            The index of the item.
        """
        item = self.asitem(item)
        args = (item, start, stop) if stop is not None else (item, start)
        return self.items.index(*args)

    def count(self, item: SequenceItemType) -> int:
        """Count the number of occurrences of a value.

        >>> ps = GeometrySequence(Point(1.0, 2.0), Point(3.0, 4.0), Point(1.0, 2.0))
        >>> ps.count(Point(1.0, 2.0))
        2

        Args:
            item: The item to count.

        Returns:
            The number of occurrences.
        """
        return self.items.count(self.asitem(item))


class PointSequence(GeometrySequence):
    """A sequence of points."""

    @property
    def points(self) -> List[Point]:
        """Return the points in the sequence.

        >>> PointSequence(Point(1.0, 2.0), Point(3.0, 4.0)).points
        [Point(x=1.0, y=2.0), Point(x=3.0, y=4.0)]
        """
        return self.items

    @classmethod
    def asitem(cls, item: PointType) -> Point:
        """Prepare the point for adding to the sequence, if necessary."""
        return Point.aspoint(item)


class LineSequence(GeometrySequence):
    """A sequence of lines."""

    @property
    def lines(self) -> List[Line]:
        """Return the lines in the sequence.

        >>> LineSequence(Line(A=1.0, B=-1.0, C=1.0), Line(A=0.0, B=1.0, C=1.0)).lines
        [Line(A=1.0, B=-1.0, C=1.0), Line(A=0.0, B=1.0, C=1.0)]
        """
        return self.items



from __future__ import annotations

from typing import Iterable, List, Tuple

import numpy as np

from .closedshape import ClosedShape
from .geometrybase import DrawAsEllipse
from .linebase import LineBase
from .point import Point, PointType


class Ellipse(ClosedShape):
    """A class representing an ellipse."""

    drawAs = DrawAsEllipse

    #: The center of the ellipse.
    center: Point
    #: The radius in the x direction.
    ra: float
    #: The radius in the y direction.
    rb: float
    #: The rotation of the ellipse in degrees.
    theta: float

    def __init__(
        self,
        center: PointType = None,
        ra: float = None,
        rb: float = None,
        theta: float = 0.0,
        A: float = None,
        B: float = None,
        C: float = None,
        D: float = None,
        E: float = None,
        F: float = None,
    ):
        """Create an ellipse, either by center, radius and rotation angle or by parameters.

        >>> Ellipse(center=Point(0.0, 0.0), ra=1.0, rb=1.0, theta=0.0)
        Ellipse(center=Point(x=0.0, y=0.0), ra=1.0, rb=1.0, theta=0.0)
        >>> Ellipse(A=1.0, B=0.0, C=1.0, D=0.0, E=0.0, F=-1.0)
        Ellipse(center=Point(x=0.0, y=0.0), ra=1.0, rb=1.0, theta=0.0)

        Args:
            center: The center of the ellipse.
            ra: The radius in the x direction.
            rb: The radius in the y direction.
            theta: The rotation of the ellipse in radians.
        """
        if center is not None and ra is not None and rb is not None and theta is not None:
            self.center = Point.aspoint(center)
            self.ra = ra
            self.rb = rb
            self.theta = theta
        elif A is not None and B is not None and C is not None and D is not None and E is not None and F is not None:
            self.ra = (
                np.sqrt(
                    2
                    * (A * E**2 + C * D**2 - B * D * E + (B**2 - 4 * A * C) * F)
                    / ((B**2 - 4 * A * C) * (np.sqrt((A - C) ** 2 + B**2) - (A + C)))
                )
                + 0.0
            )
            self.rb = (
                np.sqrt(
                    2
                    * (A * E**2 + C * D**2 - B * D * E + (B**2 - 4 * A * C) * F)
                    / ((B**2 - 4 * A * C) * (-np.sqrt((A - C) ** 2 + B**2) - (A + C)))
                )
                + 0.0
            )
            self.theta = np.arctan2(C - A - np.sqrt((A - C) ** 2 + B**2), B) + 0.0
            x0 = (2 * C * D - B * E) / (B**2 - 4 * A * C) + 0.0
            y0 = (2 * A * E - B * D) / (B**2 - 4 * A * C) + 0.0
            self.center = Point(x0, y0)
        else:
            raise ValueError("Invalid arguments.")

    def __repr__(self):
        """Return a string representation of the ellipse.

        >>> repr(Ellipse(center=Point(0.0, 0.0), ra=1.0, rb=1.0, theta=0.0))
        'Ellipse(center=Point(x=0.0, y=0.0), ra=1.0, rb=1.0, theta=0.0)'
        """
        return f"{self.__class__.__name__}(center={self.center}, ra={self.ra}, rb={self.rb}, theta={self.theta})"

    def __eq__(self, other: "Ellipse") -> bool:
        """Check if two ellipses are equal.

        >>> Ellipse(center=Point(0.0, 0.0), ra=1.0, rb=1.0) == Ellipse(center=Point(0.0, 0.0), ra=1.0, rb=1.0)
        True
        >>> Ellipse(center=Point(0.0, 0.0), ra=1.0, rb=1.0) == Ellipse(center=Point(0.0, 0.0), ra=1.0, rb=2.0)
        False

        Args:
            other: The other ellipse.
        """
        return self.center == other.center and self.ra == other.ra and self.rb == other.rb and self.theta == other.theta

    @property
    def boundaries(self) -> List[LineBase]:
        raise NotImplementedError

    def containsPoint(self, point: PointType | Iterable[Point]) -> bool:
        """Check if a point is inside the ellipse.

        >>> Point(0.5, 0.5) in Ellipse(center=Point(0.0, 0.0), ra=1.0, rb=1.0)
        True
        >>> Point(0.6, 0.8) in Ellipse(center=Point(0.0, 0.0), ra=1.0, rb=1.0)
        True

        Args:
            point: The point.
        """
        if isinstance(point, Iterable) and isinstance(tuple(point)[0], Point):
            return all(self.contains(p) for p in point)
        point = Point.aspoint(point)
        x, y = point.x, point.y
        A, B, C, D, E, F = self.coefficients
        return A * x**2 + B * x * y + C * y**2 + D * x + E * y + F <= 0

    def containsLine(self, line: LineBase | Iterable[LineBase]) -> bool:
        raise NotImplementedError

    @property
    def x0(self) -> float:
        """Return the x coordinate of the left side of the ellipse."""
        return self.center.x

    @property
    def y0(self) -> float:
        """Return the y coordinate of the top of the ellipse."""
        return self.center.y

    @property
    def A(self) -> float:
        """Returns A parameter of the ellipse equation."""
        return self.ra**2 * np.sin(self.theta) ** 2 + self.rb**2 * np.cos(self.theta) ** 2

    @property
    def B(self) -> float:
        """Returns B parameter of the ellipse equation."""
        return 2 * (self.rb**2 - self.ra**2) * np.sin(self.theta) * np.cos(self.theta)

    @property
    def C(self) -> float:
        """Returns C parameter of the ellipse equation."""
        return self.ra**2 * np.cos(self.theta) ** 2 + self.rb**2 * np.sin(self.theta) ** 2

    @property
    def D(self) -> float:
        """Returns D parameter of the ellipse equation."""
        return -2 * self.A * self.x0 - self.B * self.y0

    @property
    def E(self) -> float:
        """Returns E parameter of the ellipse equation."""
        return -self.B * self.x0 - 2 * self.C * self.y0

    @property
    def F(self) -> float:
        """Returns F parameter of the ellipse equation."""
        return self.A * self.x0**2 + self.B * self.x0 * self.y0 + self.C * self.y0**2 - self.ra**2 * self.rb**2

    @property
    def coefficients(self) -> Tuple[float, float, float, float, float, float]:
        """Return the coefficients of the ellipse equation.

        Returns:
            The parameters of the ellipse equation.
        """
        return self.A, self.B, self.C, self.D, self.E, self.F

    @property
    def maxsize(self) -> float:
        """Return the maximum size of the ellipse.

        >>> Ellipse(center=Point(0.0, 0.0), ra=1.0, rb=2.0).maxsize
        4.0
        """
        return max(self.ra, self.rb) * 2

    @property
    def area(self) -> float:
        """Return the area of the ellipse.

        >>> assert Ellipse(center=Point(0.0, 0.0), ra=1.0, rb=1.0).area == np.pi
        """
        return np.pi * self.ra * self.rb

    @property
    def perimeter(self) -> float:
        """Return the perimeter of the ellipse.

        >>> assert Ellipse(center=Point(0.0, 0.0), ra=1.0, rb=1.0).perimeter == 2 * np.pi
        """
        return 2 * np.pi * np.sqrt((self.ra**2 + self.rb**2) / 2)

    @property
    def centroid(self) -> Point:
        """Return the centroid of the ellipse.

        >>> Ellipse(center=Point(0.0, 0.0), ra=1.0, rb=1.0).centroid
        Point(x=0.0, y=0.0)
        """
        return self.center

    @property
    def bbox(self) -> Tuple[Point, Point]:
        """Return the bounding box of the ellipse."""
        raise NotImplementedError

    @property
    def drawArgs(self) -> Tuple[float, float, float, float]:
        return self.x0, self.y0, self.ra, self.rb



from __future__ import annotations

from svgpathtools.path import Path

from .filterbase import FilterBase


class SegmentNumberFilter(FilterBase):
    """Filter paths based on the number of segments."""

    #: Minimum number of segments in a path.
    min_segments: int

    def __init__(self, min_segments: int = 4, *, enabled: bool = True, tolerance: float = 1e-6):
        super().__init__(enabled=enabled, tolerance=tolerance)
        self.min_segments = min_segments

    def accept(self, path: Path) -> bool:
        return not self.enabled or len(path) >= self.min_segments



from .closedpath import ClosedPathFilter
from .custom import CustomFilter
from .filterbase import FilterBase
from .rectangle import RectangleRangeFilter
from .segmentnumber import SegmentNumberFilter
from .specialline import HorizontalLineFilter, VerticalLineFilter

__all__ = [
    "ClosedPathFilter",
    "CustomFilter",
    "FilterBase",
    "RectangleRangeFilter",
    "SegmentNumberFilter",
    "HorizontalLineFilter",
    "VerticalLineFilter",
]



from __future__ import annotations

from typing import Callable

import numpy as np
from svgpathtools.path import Path
from svgpathtools.svg_to_paths import svg2paths

from .filters import (
    ClosedPathFilter,
    CustomFilter,
    FilterBase,
    HorizontalLineFilter,
    RectangleRangeFilter,
    SegmentNumberFilter,
    VerticalLineFilter,
)
from .geometry import PathSequence
from .geometry.csys import CoordinateSystem


class SvgPathParser:
    """Parse an SVG file and return the paths."""

    #: Path to the svg file
    svgfile: str
    #: Filters
    filters: list[FilterBase]
    #: coordinate system
    csys: CoordinateSystem

    def __init__(
        self,
        svgfile: str,
        xrange: tuple[float, float] = (-np.inf, np.inf),
        yrange: tuple[float, float] = (-np.inf, np.inf),
        min_segments: int = 4,
        drop_horizontal_lines: bool = True,
        drop_vertical_lines: bool = True,
        drop_closed_paths: bool = True,
        custom_filter: Callable[[Path], bool] | FilterBase = None,
        tolerance: float = 1e-6,
    ):
        """Constructor of the SvgPathParser class.

        Args:
            svgfile: Path to the svg file.
            xrange: Range of x values to include.
            yrange: Range of y values to include.
            min_segments: Minimum number of segments in a path.
            drop_horizontal_lines: Whether to drop horizontal lines.
            drop_vertical_lines: Whether to drop closed paths.
            drop_closed_paths: Whether to drop closed paths.
            custom_filter: Custom filter for the paths.
            tolerance: Tolerance for determining if a path is a horizontal or vertical line.
        """
        self.svgfile = svgfile
        self.filters = [
            RectangleRangeFilter(xrange=xrange, yrange=yrange),
            SegmentNumberFilter(min_segments=min_segments),
            HorizontalLineFilter(enabled=drop_horizontal_lines, tolerance=tolerance),
            VerticalLineFilter(enabled=drop_vertical_lines, tolerance=tolerance),
            ClosedPathFilter(enabled=drop_closed_paths),
        ]
        self.addFilter(custom_filter)
        self.csys = CoordinateSystem()

    def addFilter(self, f: Callable[[Path], bool] | FilterBase):
        """Add a custom filter to the parser.

        Args:
            f: Custom filter for the paths.
        """
        if not f:
            return
        self.filters.append(f if isinstance(f, FilterBase) else CustomFilter(f))

    def parse(self) -> PathSequence:
        """Parse the paths from the svg file.

        Returns:
            A list of paths.
        """

        def filtered(pts: list[Path]) -> list[Path]:
            return [pt for pt in pts if all(f.accept(pt) for f in self.filters)]

        paths, atts = svg2paths(self.svgfile)
        paths = PathSequence.fromSvgPathToolsPathSequence(filtered(paths)).transformed(self.csys)
        return paths



from __future__ import annotations

from pathlib import Path
from re import error as RegexError  # noqa


def _get_version():
    root = Path(__file__).resolve().parents[1]
    if (root / ".git").exists() and not (root / ".git/shallow").exists():
        try:
            import setuptools_scm  # noqa

            return setuptools_scm.get_version(root=str(root))
        except (ImportError, RegexError, LookupError):
            pass
    try:
        from ._version import version

        return version
    except ImportError:
        return "0.0.0-unknown"


__version__: str = _get_version()
__semver__: str = __version__.split("+")[0]



from qtpy.QtCore import Qt
from qtpy.QtWidgets import QMainWindow

from ..geometry import Circle, Point, Polygon
from .graphicsview import GraphicsView


class SvgDataGrabberMainWindow(QMainWindow):
    """A custom main window for the data grabber application."""

    def __init__(self, parent=None):
        super(SvgDataGrabberMainWindow, self).__init__(parent)
        self.view = GraphicsView(self)
        self.setCentralWidget(self.view)

        self.resize(800, 600)
        self.setWindowTitle("SVG Data Grabber")

        # add test geometries
        self.view.addPrimitive(Polygon(Point(0.0, 0.0), Point(100.0, 0.0), Point(100.0, 100.0), Point(0.0, 100.0)))
        self.view.addPrimitive(Circle(center=Point(50.0, 50.0), r=50.0))
        self.view.redraw(Qt.blue, Qt.red)



from __future__ import annotations

from typing import Iterable

from .point import Point
from .straightline import ExtendedLineSegment


class Axis(ExtendedLineSegment):
    pass


class XAxis(Axis):
    #: X value of the first point.
    xstart: float
    #: X value of the second point.
    xend: float
    #: Y value of the axis.
    y: float

    def __init__(
        self,
        start: Point | Iterable[float] | complex = Point(0.0, 0.0),
        end: Point | Iterable[float] | complex = Point(1.0, 0.0),
        xstart: float = 0.0,
        xend: float = 1.0,
        y: float = 0.0,
    ):
        super().__init__(start=start, end=end)
        self.setup(start=start, end=end, xstart=xstart, xend=xend, y=y)

    def setup(
        self,
        start: Point | Iterable[float] | complex = Point(0.0, 0.0),
        end: Point | Iterable[float] | complex = Point(1.0, 0.0),
        xstart: float = 0.0,
        xend: float = 1.0,
        y: float = 0.0,
    ):
        """Set up the axis.

        Args:
            start: The first point to create the line.
            end: The second point to create the line.
            xstart: X value of the first point.
            xend: X value of the second point.
            y: Y value of the axis.
        """
        start, end = Point.aspoint(start), Point.aspoint(end)
        self.A, self.B, self.C = self.coefficientsFromTwoPoints(start, end)
        self.start, self.end = start, end
        self.xstart, self.xend = xstart, xend
        self.y = y


class YAxis(Axis):
    #: Y value of the first point.
    ystart: float
    #: Y value of the second point.
    yend: float
    #: X value of the axis.
    x: float

    def __init__(
        self,
        start: Point | Iterable[float] | complex = Point(0.0, 0.0),
        end: Point | Iterable[float] | complex = Point(0.0, 1.0),
        ystart: float = 0.0,
        yend: float = 1.0,
        x: float = 0.0,
    ):
        super().__init__(start=start, end=end)
        self.setup(start=start, end=end, ystart=ystart, yend=yend, x=x)

    def setup(
        self,
        start: Point | Iterable[float] | complex = Point(0.0, 0.0),
        end: Point | Iterable[float] | complex = Point(0.0, 1.0),
        ystart: float = 0.0,
        yend: float = 1.0,
        x: float = 0.0,
    ):
        """Set up the axis.

        Args:
            start: The first point to create the line.
            end: The second point to create the line.
            ystart: Y value of the first point.
            yend: Y value of the second point.
            x: X value of the axis.
        """
        start, end = Point.aspoint(start), Point.aspoint(end)
        self.A, self.B, self.C = self.coefficientsFromTwoPoints(start, end)
        self.start, self.end = start, end
        self.ystart, self.yend = ystart, yend
        self.x = x


class CoordinateSystem:
    #: The x-axis of the coordinate system.
    xaxis: XAxis
    #: The y-axis of the coordinate system.
    yaxis: YAxis

    def __init__(self):
        self.xaxis = XAxis()
        self.yaxis = YAxis()

    def transform(self, p: Point | Iterable[float] | complex) -> Point:
        """Transform the coordinate to the coordinate system.

        >>> csys = CoordinateSystem()
        >>> csys.transform(Point(0.0, 0.0))
        Point(x=0.0, y=0.0)
        >>> csys.transform(Point(1.0, 1.0))
        Point(x=1.0, y=1.0)

        >>> csys = CoordinateSystem()
        >>> csys.setup_xaxis(start=(0.0, 0.0), end=(2.0, 0.0), xstart=0.0, xend=1.0, y=0.0, check=False)
        >>> csys.setup_yaxis(start=(0.0, 0.0), end=(0.0, 2.0), ystart=0.0, yend=1.0, x=0.0, check=True)
        >>> csys.transform(Point(2.0, 0.0))
        Point(x=1.0, y=0.0)
        >>> csys.transform(Point(0.0, 2.0))
        Point(x=0.0, y=1.0)

        >>> csys = CoordinateSystem()
        >>> csys.setup_xaxis(start=(1.0, 0.0), end=(2.0, 0.0), xstart=0.0, xend=1.0, y=0.0, check=False)
        >>> csys.setup_yaxis(start=(0.0, 1.0), end=(0.0, 2.0), ystart=0.0, yend=1.0, x=0.0, perpendicular=True)
        >>> csys.transform(Point(2.0, 1.0))
        Point(x=1.0, y=0.0)
        >>> csys.transform(Point(1.0, 2.0))
        Point(x=0.0, y=1.0)

        >>> csys = CoordinateSystem()
        >>> csys.setup_xaxis(start=(0.0, 0.0), end=(1.0, 0.0), xstart=0.0, xend=1.0, y=0.0)
        >>> csys.setup_yaxis(start=(0.0, 0.0), end=(1.0, 1.0), ystart=0.0, yend=1.0, x=0.0, check=True)
        Traceback (most recent call last):
        ...
        ValueError: The x-axis and y-axis must be perpendicular.
        >>> csys.setup_yaxis(start=(0.0, 0.0), end=(0.0, 1.0), ystart=0.0, yend=1.0, x=0.0)
        >>> csys.setup_xaxis(start=(0.0, 0.0), end=(1.0, 1.0), xstart=0.0, xend=1.0, y=0.0, check=True)
        Traceback (most recent call last):
        ...
        ValueError: The x-axis and y-axis must be perpendicular.

        Args:
            p: The point to convert.

        Returns:
            The coordinate in the coordinate system.
        """
        xp = self.yaxis.parallel(p).intersect(self.xaxis)
        yp = self.xaxis.parallel(p).intersect(self.yaxis)
        rx = (xp.x - self.xaxis.start.x) / (self.xaxis.end.x - self.xaxis.start.x)
        ry = (yp.y - self.yaxis.start.y) / (self.yaxis.end.y - self.yaxis.start.y)
        x = self.xaxis.xstart + rx * (self.xaxis.xend - self.xaxis.xstart)
        y = self.yaxis.ystart + ry * (self.yaxis.yend - self.yaxis.ystart)
        x, y = round(x + 0.0, 10), round(y + 0.0, 10)
        return Point(x, y)

    def setup_xaxis(
        self,
        start: Point | Iterable[float] | complex = Point(0.0, 0.0),
        end: Point | Iterable[float] | complex = Point(1.0, 0.0),
        xstart: float = 0.0,
        xend: float = 1.0,
        y: float = 0.0,
        perpendicular: bool = False,
        check: bool = False,
    ):
        """Set up the x-axis.

        Args:
            start: The first point to create the line.
            end: The second point to create the line.
            xstart: X value of the first point.
            xend: X value of the second point.
            y: Y value of the axis.
            perpendicular: If the axis should be perpendicular to the y-axis. If True, the y value of `end` will be
                           updated.
            check: Check if the axis is perpendicular to the y-axis.
        """
        start, end = Point.aspoint(start), Point.aspoint(end)
        if perpendicular:
            end = self.yaxis.perpendicular(start).intersect(self.yaxis.parallel(end))
        self.xaxis.setup(start=start, end=end, xstart=xstart, xend=xend, y=y)
        if check and not perpendicular and not self.xaxis.isPerpendicular(self.yaxis):
            raise ValueError("The x-axis and y-axis must be perpendicular.")

    def setup_yaxis(
        self,
        start: Point | Iterable[float] | complex = Point(0.0, 0.0),
        end: Point | Iterable[float] | complex = Point(0.0, 1.0),
        ystart: float = 0.0,
        yend: float = 1.0,
        x: float = 0.0,
        perpendicular: bool = False,
        check: bool = False,
    ):
        """Set up the y-axis.

        Args:
            start: The first point to create the line.
            end: The second point to create the line.
            ystart: Y value of the first point.
            yend: Y value of the second point.
            x: X value of the axis.
            perpendicular: If the axis should be perpendicular to the y-axis. If True, the x value of `end` will be
                           updated.
            check: Check if the axis is perpendicular to the y-axis.
        """
        start, end = Point.aspoint(start), Point.aspoint(end)
        if perpendicular:
            end = self.xaxis.perpendicular(start).intersect(self.xaxis.parallel(end))
        self.yaxis.setup(start=start, end=end, ystart=ystart, yend=yend, x=x)
        if check and not perpendicular and not self.yaxis.isPerpendicular(self.xaxis):
            raise ValueError("The x-axis and y-axis must be perpendicular.")



from typing import Tuple

from .ellipse import Ellipse
from .point import Point, PointType


class Circle(Ellipse):
    """A class representing a circle."""

    #: The radius of the circle.
    _r: float

    def __init__(self, center: PointType, r: float):
        """Create a circle, either by center and radius or by parameters.

        >>> Circle(center=Point(0.0, 0.0), r=1.0)
        Circle(center=Point(x=0.0, y=0.0), r=1.0)

        Args:
            center: The center of the circle.
            r: The radius of the circle.
        """
        super().__init__(center=center, ra=r, rb=r, theta=0.0)
        self._r = r

    def __repr__(self) -> str:
        """Get the string representation of the circle.

        >>> repr(Circle(center=Point(0.0, 0.0), r=1.0))
        'Circle(center=Point(x=0.0, y=0.0), r=1.0)'
        """
        return f"{self.__class__.__name__}(center={self.center}, r={self.r})"

    @property
    def r(self):
        """Get the radius of the circle.

        >>> Circle(center=Point(0.0, 0.0), r=1.0).r
        1.0
        """
        return self._r

    @r.setter
    def r(self, value):
        """Set the radius of the circle.

        >>> circle = Circle(center=Point(0.0, 0.0), r=1.0)
        >>> circle.r = 2.0
        >>> circle.r
        2.0
        """
        self._r, self.ra, self.rb = value, value, value

    def bbox(self) -> Tuple[Point, Point]:
        """Get the bounding box of the circle.

        >>> Circle(center=Point(0.0, 0.0), r=1.0).bbox()
        (Point(x=-1.0, y=-1.0), Point(x=1.0, y=1.0))

        Returns:
            The bounding box of the circle.
        """
        return (
            Point(x=self.center.x - self.r, y=self.center.y - self.r),
            Point(x=self.center.x + self.r, y=self.center.y + self.r),
        )



from typing import Union

from qtpy.QtCore import Qt
from qtpy.QtGui import QBrush, QColor, QGradient, QPen
from qtpy.QtWidgets import QGraphicsEllipseItem, QGraphicsLineItem, QGraphicsPolygonItem

QPenType = Union[QPen, QColor, Qt.GlobalColor, QGradient]
QBrushType = Union[QBrush, QColor, Qt.GlobalColor, QGradient]
QGraphicsItemType = Union[QGraphicsLineItem, QGraphicsPolygonItem, QGraphicsEllipseItem]



from __future__ import annotations

from typing import Iterable, List, Tuple

import numpy as np
from svgpathtools import Arc as SvgPathToolsArc

from .ellipse import Ellipse
from .linebase import CurveLineBase, LineBase
from .point import Point, PointType


class Arc(CurveLineBase, Ellipse):
    """A class representing an arc."""

    #: The start point of the arc.
    start: Point
    #: The end point of the arc.
    end: Point
    #: Whether the arc is the large arc determined by the start and end points.
    large_arc: bool
    #: Whether the arc is the sweep arc determined by the start and end points.
    sweep: bool
    #: Whether to automatically scale the radius to fit the start and end points.
    autoscale_radius: bool

    def __init__(
        self,
        center: PointType,
        ra: float,
        rb: float,
        theta: float,
        start: PointType,
        end: PointType,
        large_arc: bool,
        sweep: bool,
        autoscale_radius: bool = True,
    ):
        """Create an arc.

        >>> Arc(center=Point(0.0, 0.0), ra=1.0, rb=1.0, theta=0.0, start=Point(1.0, 0.0), end=Point(0.0, 1.0), large_arc=False, sweep=True)
        Arc(center=Point(x=0.0, y=0.0), ra=1.0, rb=1.0, theta=0.0, start=Point(x=1.0, y=0.0), end=Point(x=0.0, y=1.0), large_arc=False, sweep=True)

        Args:
            center: The center of the arc.
            ra: The radius in the x direction.
            rb: The radius in the y direction.
            theta: The rotation of the arc in radians.
            start: The start point of the arc.
            end: The end point of the arc.
            large_arc: Whether the arc is the large arc determined by the start and end points.
            sweep: Whether the arc is the sweep arc determined by the start and end points.
            autoscale_radius: Whether to automatically scale the radius to fit the start and end points.
        """
        Ellipse.__init__(self, center, ra, rb, theta)
        self.start = Point.aspoint(start)
        self.end = Point.aspoint(end)
        self.large_arc = large_arc
        self.sweep = sweep
        self.autoscale_radius = autoscale_radius

    def __repr__(self) -> str:
        """Return a string representation of the arc."""
        return (
            f"{self.__class__.__name__}(center={self.center}, ra={self.ra}, rb={self.rb}, theta={self.theta}, "
            f"start={self.start}, end={self.end}, large_arc={self.large_arc}, sweep={self.sweep})"
        )

    def __eq__(self, other: "Arc") -> bool:
        """Return whether the arc is equal to another arc.

        >>> arc1 = Arc(Point(0.0, 0.0), 1.0, 1.0, 0.0, Point(1.0, 0.0), Point(0.0, 1.0), False, True)
        >>> arc2 = Arc(Point(0.0, 0.0), 1.0, 1.0, 0.0, Point(1.0, 0.0), Point(0.0, 1.0), False, True)
        >>> arc1 == arc2
        True
        """
        return (
            Ellipse.__eq__(self, other)
            and self.start == other.start
            and self.end == other.end
            and self.large_arc == other.large_arc
            and self.sweep == other.sweep
        )

    def __iter__(self):
        """Return an iterator over the arc.

        >>> arc = Arc(Point(0.0, 0.0), 1.0, 1.0, 0.0, Point(1.0, 0.0), Point(0.0, 1.0), False, True)
        >>> list(arc)
        [Point(x=1.0, y=0.0), Point(x=0.0, y=1.0)]
        """
        yield self.start
        yield self.end

    @classmethod
    def fromSvgPathToolsArc(cls, arc: SvgPathToolsArc) -> Arc:
        """Create an arc from a svgpathtools Arc."""
        return cls(
            center=Point(arc.center),
            ra=arc.radius.real,
            rb=arc.radius.imag,
            theta=np.deg2rad(arc.rotation),
            start=Point(arc.start),
            end=Point(arc.end),
            large_arc=arc.large_arc,
            sweep=arc.sweep,
        )

    @property
    def bbox(self) -> Tuple[Point, Point]:
        raise NotImplementedError

    def containsLine(self, line: LineBase | Iterable[LineBase]) -> bool:
        raise TypeError("Arc does not contain a line.")

    @property
    def boundaries(self) -> List[LineBase]:
        raise NotImplementedError



import numpy as np

from .point import Point, PointType  # noqa
from .polygon import Polygon
from .straightline import LineSegment  # noqa


class Quadrilateral(Polygon):
    """A class representing a quadrilateral."""

    def __init__(self, *points: PointType):
        if len(points) != 4:
            raise ValueError("A quadrilateral must have exactly four points.")
        super().__init__(*points)

    def check(self):
        """Check if the quadrilateral is convex and complex.

        >>> Quadrilateral(Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0), Point(0.0, 1.0)).check()
        >>> Quadrilateral(Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0), Point(0.0, -1.0)).check()
        Traceback (most recent call last):
        ...
        AssertionError: The polygon must be simple.
        """
        super().check()
        assert self.isValid, "The quadrilateral must be convex."


class Kite(Quadrilateral):
    """A class representing a kite."""

    @property
    def isKite(self):
        """Check if the quadrilateral is a kite."""
        (l1, l2, l3, l4) = (edge.length for edge in self.edges)  # type: float
        return np.allclose([l1, l3], [l2, l4]) or np.allclose([l2, l1], [l3, l4])

    def check(self):
        """Check if the quadrilateral is a kite.

        >>> Kite(Point(1.0, 0.0), Point(0.0, 1.0), Point(-1.0, 0.0), Point(0.0, -2.0)).check()
        >>> Kite(Point(2.0, 0.0), Point(0.0, 1.0), Point(-1.0, 0.0), Point(0.0, -2.0)).check()
        Traceback (most recent call last):
        ...
        AssertionError: Two pairs of adjacent sides must be of equal length.
        """
        super().check()
        assert self.isKite, "Two pairs of adjacent sides must be of equal length."


class RightKite(Kite):
    """A class representing a right kite."""

    @property
    def isRightKite(self):
        """Check if the quadrilateral is a right kite."""
        e1, e2, e3, e4 = self.edges  # type: LineSegment
        l1, l2, l3, l4 = e1.length, e2.length, e3.length, e4.length
        return (np.allclose([l1, l3], [l2, l4]) and e1.isPerpendicular(e4)) or (
            np.allclose([l2, l1], [l3, l4]) and e1.isPerpendicular(e2)
        )

    def check(self):
        """Check if the quadrilateral is a right kite.

        >>> RightKite(Point(1.0, 0.0), Point(0.0, 1.0), Point(-1.0, 0.0), Point(0.0, -2.0)).check()
        Traceback (most recent call last):
        ...
        AssertionError: A kite must be inscribed in a circle.
        >>> RightKite(Point(3.0, 0.0), Point(0.0, 4.0), Point(-3.0, 0.0), Point(0.0, -2.25)).check()
        """
        super().check()
        assert self.isRightKite, "A kite must be inscribed in a circle."


class Rhombus(Kite):
    """A class representing a rhombus."""

    @property
    def isRhombus(self):
        """Check if the quadrilateral is a rhombus."""
        (l1, l2, l3, l4) = (edge.length for edge in self.edges)  # type: float
        return np.allclose([l2, l3, l4], l1)

    def check(self):
        """Check if the quadrilateral is a rhombus.

        >>> Rhombus(Point(1.0, 0.0), Point(0.0, 2.0), Point(-1.0, 0.0), Point(0.0, -2.0)).check()
        >>> Rhombus(Point(2.0, 0.0), Point(0.0, 2.0), Point(-1.0, 0.0), Point(0.0, -2.0)).check()
        Traceback (most recent call last):
        ...
        AssertionError: All sides must be of equal length.
        """
        super().check()
        assert self.isRhombus, "All sides must be of equal length."


class Trapezoid(Quadrilateral):
    """A class representing a trapezoid."""

    @property
    def isTrapezoid(self):
        """Check if the quadrilateral is a trapezoid."""
        e1, e2, e3, e4 = self.edges  # type: LineSegment
        return e1.isParallel(e3) or e2.isParallel(e4)

    def check(self):
        """Check if the quadrilateral is a trapezoid.

        >>> Trapezoid(Point(2.0, 0.0), Point(1.0, 1.0), Point(-1.0, 1.0), Point(-3.0, 0.0)).check()
        >>> Trapezoid(Point(2.0, 0.0), Point(1.0, 1.0), Point(-1.0, 2.0), Point(-3.0, 0.0)).check()
        Traceback (most recent call last):
        ...
        AssertionError: Two pairs of adjacent sides must be parallel.
        """
        super().check()
        assert self.isTrapezoid, "Two pairs of adjacent sides must be parallel."


class IsoscelesTrapezoid(Trapezoid):
    """A class representing an isosceles trapezoid."""

    @property
    def isIsoscelesTrapezoid(self):
        """Check if the quadrilateral is an isosceles trapezoid."""
        #: https://en.wikipedia.org/wiki/Isosceles_trapezoid
        e1, e2, e3, e4 = self.edges  # type: LineSegment
        l1, l2, l3, l4 = e1.length, e2.length, e3.length, e4.length
        return (e1.isParallel(e3) and np.allclose(l2, l4)) or (e2.isParallel(e4) and np.allclose(l1, l3))

    def check(self):
        """Check if the quadrilateral is an isosceles trapezoid.

        >>> IsoscelesTrapezoid(Point(2.0, 0.0), Point(1.0, 1.0), Point(-1.0, 1.0), Point(-2.0, 0.0)).check()
        >>> IsoscelesTrapezoid(Point(2.0, 0.0), Point(1.0, 1.0), Point(-1.0, 1.0), Point(-3.0, 0.0)).check()
        Traceback (most recent call last):
        ...
        AssertionError: Two pairs of adjacent sides must be parallel and of equal length.
        """
        super().check()
        assert self.isIsoscelesTrapezoid, "Two pairs of adjacent sides must be parallel and of equal length."


class Parallelogram(Quadrilateral):
    """A class representing a parallelogram."""

    @property
    def isParallelogram(self):
        """Check if the quadrilateral is a parallelogram."""
        e1, e2, e3, e4 = self.edges  # type: LineSegment
        return e1.isParallel(e3) and e2.isParallel(e4)

    def check(self):
        """Check if the quadrilateral is a parallelogram.

        >>> Parallelogram(Point(0.0, 0.0), Point(2.0, 0.0), Point(3.0, 1.0), Point(1.0, 1.0)).check()
        >>> Parallelogram(Point(-1.0, 0.0), Point(2.0, 0.0), Point(3.0, 1.0), Point(1.0, 1.0)).check()
        Traceback (most recent call last):
        ...
        AssertionError: Two pairs of adjacent sides must be parallel.
        """
        super().check()
        assert self.isParallelogram, "Two pairs of adjacent sides must be parallel."


class Rectangle(Parallelogram, IsoscelesTrapezoid):
    """A class representing a rectangle."""

    @property
    def isRectangle(self):
        """Check if the quadrilateral is a rectangle."""
        e1, e2, e3, e4 = self.edges  # type: LineSegment
        return e1.isPerpendicular(e2)

    def check(self):
        """Check if the quadrilateral is a rectangle.

        >>> Rectangle(Point(0.0, 0.0), Point(2.0, 0.0), Point(2.0, 1.0), Point(0.0, 1.0)).check()
        >>> Rectangle(Point(0.0, 0.0), Point(2.0, 0.0), Point(3.0, 1.0), Point(1.0, 1.0)).check()
        Traceback (most recent call last):
        ...
        AssertionError: The quadrilateral must be a right parallelogram.
        """
        super().check()
        assert self.isRectangle, "The quadrilateral must be a right parallelogram."


class Square(Rectangle, Rhombus, Parallelogram, RightKite):
    """A class representing a square."""

    @property
    def isSquare(self):
        """Check if the quadrilateral is a square."""
        #: https://en.wikipedia.org/wiki/Square
        (l1, l2, l3, l4) = (edge.length for edge in self.edges)  # type: float
        return np.allclose([l1, l2, l3, l4], l1)

    def check(self):
        """Check if the quadrilateral is a square.

        >>> Square(Point(0.0, 0.0), Point(2.0, 0.0), Point(2.0, 2.0), Point(0.0, 2.0)).check()
        >>> Square(Point(0.0, 0.0), Point(2.0, 0.0), Point(2.0, 1.0), Point(0.0, 1.0)).check()
        Traceback (most recent call last):
        ...
        AssertionError: Two pairs of adjacent sides must be of equal length.
        """
        super().check()
        assert self.isSquare, "All sides must be of equal length."



from __future__ import annotations

from typing import Iterable, Tuple, Union

import numpy as np
from qtpy.QtCore import QLineF

from .geometrybase import DrawAsEllipse, GeometryBase

PointType = Union["Point", Iterable[float], complex]


class Point(GeometryBase):
    """A class representing a point in 2D space."""

    drawAs = DrawAsEllipse

    #: The x coordinate of the point.
    x: float
    #: The y coordinate of the point.
    y: float

    def __init__(self, *args: float | PointType, x: float = None, y: float = None):
        """Initialize a point.

        >>> Point(1.0, 2.0)
        Point(x=1.0, y=2.0)
        >>> Point([1.0, 2.0])
        Point(x=1.0, y=2.0)
        >>> Point(complex(1.0, 2.0))
        Point(x=1.0, y=2.0)
        >>> Point(Point(1.0, 2.0))
        Point(x=1.0, y=2.0)
        >>> Point(x=1.0, y=2.0)
        Point(x=1.0, y=2.0)
        >>> Point(1.0)
        Traceback (most recent call last):
        ...
        ValueError: The arguments must be two floats, a complex number or an iterable of two floats.

        Args:
            args: two floats, a complex number or an iterable of two floats.
        """
        if len(args) == 1 and isinstance(args[0], complex):
            self.x, self.y = args[0].real, args[0].imag
        elif len(args) == 1 and isinstance(args[0], Iterable) and len(tuple(args[0])) == 2:
            self.x, self.y = args[0]
        elif len(args) == 2:
            self.x, self.y = args
        elif len(args) == 0 and x is not None and y is not None:
            self.x, self.y = x, y
        else:
            raise ValueError("The arguments must be two floats, a complex number or an iterable of two floats.")

    def __repr__(self) -> str:
        """Return a string representation of the point.

        >>> repr(Point(1.0, 2.0))
        'Point(x=1.0, y=2.0)'
        """
        x, y = round(self.x, 10), round(self.y, 10)
        return f"{self.__class__.__name__}(x={x}, y={y})"

    def __iter__(self):
        """Iterate over the coordinates of the point.

        >>> point = Point(1.0, 2.0)
        >>> x, y = point
        >>> assert x == 1.0
        >>> assert y == 2.0
        """
        yield self.x
        yield self.y

    def __neg__(self) -> Point:
        """Return the negative of the point.

        >>> point = -Point(1.0, 2.0)
        >>> assert point.x == -1.0
        >>> assert point.y == -2.0
        """
        return self.__class__(-self.x, -self.y)

    def __eq__(self, other: PointType) -> bool:
        """Check if two points are equal.

        >>> Point(1.0, 2.0) == Point(1.0, 2.0)
        True
        >>> Point(1.0, 2.0) != Point(1.0, 2.0)
        False
        >>> Point(1.0, 2.0) == Point(1.0, 3.0)
        False
        >>> Point(1.0, 2.0) != Point(1.0, 3.0)
        True

        Args:
            other: The other point.
        """
        other = self.aspoint(other)
        return np.allclose([self.x, self.y], [other.x, other.y], atol=self.tolerance)

    def __add__(self, other: PointType) -> Point:
        """Add a point or vector to the point.

        >>> Point(1.0, 2.0) + Point(3.0, 4.0)
        Point(x=4.0, y=6.0)

        Args:
            other: The other point.
        """
        other = self.aspoint(other)
        return Point(self.x + other.x, self.y + other.y)

    def __sub__(self, other: PointType) -> Point:
        """Subtract a point or vector from the point.

        >>> Point(1.0, 2.0) - Point(3.0, 4.0)
        Point(x=-2.0, y=-2.0)

        Args:
            other: The other point.
        """
        other = self.aspoint(other)
        return Point(self.x - other.x, self.y - other.y)

    def __mul__(self, other: float) -> Point:
        """Multiply the point by a scalar.

        >>> Point(1.0, 2.0) * 2.0
        Point(x=2.0, y=4.0)

        Args:
            other: A scalar value.
        """
        return Point(self.x * other, self.y * other)

    def __truediv__(self, other: float) -> Point:
        """Divide the point by a scalar.

        >>> Point(1.0, 2.0) / 2.0
        Point(x=0.5, y=1.0)

        Args:
            other: A scalar value.
        """
        return Point(self.x / other, self.y / other)

    def __abs__(self) -> float:
        """Calculate the magnitude of the vector.

        >>> assert abs(Point(3.0, 4.0)) == 5.0

        Returns:
            The magnitude of the vector.
        """
        return np.linalg.norm(self.array)

    @classmethod
    def aspoint(cls, p: PointType) -> Point:
        """Convert a point to a Point object.

        >>> Point.aspoint(Point(1.0, 2.0))
        Point(x=1.0, y=2.0)
        >>> Point.aspoint((1.0, 2.0))
        Point(x=1.0, y=2.0)
        >>> Point.aspoint(1.0 + 2.0j)
        Point(x=1.0, y=2.0)
        >>> Point.aspoint(1.0)
        Traceback (most recent call last):
        ...
        TypeError: Cannot convert 1.0 to a Point object.

        Args:
            p: The point to convert.

        Returns:
            The converted point.
        """
        if isinstance(p, cls):
            return p
        elif isinstance(p, complex):
            return cls(p.real, p.imag)
        elif isinstance(p, Iterable):
            return cls(*tuple(p))
        else:
            raise TypeError(f"Cannot convert {p} to a {cls.__name__} object.")

    @property
    def maxsize(self) -> float:
        """Return the maximum size of the point.

        >>> Point(1.0, 2.0).maxsize
        0.0
        """
        return 0.0

    def distance(self, other: PointType) -> float:
        """Calculate the distance between two points.

        >>> assert Point(1.0, 2.0).distance(Point(4.0, 6.0)) == 5.0

        Args:
            other: The other point.

        Returns:
            The distance between the points.
        """
        other = self.aspoint(other)
        return np.linalg.norm(np.array([self.x, self.y]) - np.array([other.x, other.y]))

    def direction(self, other: PointType) -> float:
        """Calculate the direction between two points.

        >>> p1 = Point(0.0, 0.0).direction(Point(1.0, 1.0))
        >>> assert np.isclose(p1, np.pi / 4.0)
        >>> p2 = Point(0.0, 0.0).direction(Point(-1.0, 1.0))
        >>> assert np.isclose(p2, 3.0 * np.pi / 4.0)
        >>> p2 = Point(0.0, 0.0).direction(Point(-1.0, -1.0))
        >>> assert np.isclose(p2, -3.0 * np.pi / 4.0)
        >>> p2 = Point(0.0, 0.0).direction(Point(1.0, -1.0))
        >>> assert np.isclose(p2, -np.pi / 4.0)

        Args:
            other: The other point.

        Returns:
            The direction between the points.
        """
        other = self.aspoint(other)
        return np.arctan2(other.y - self.y, other.x - self.x)

    def vector(self, other: PointType) -> Vector:
        """Calculate the vector between two points.

        >>> Point(1.0, 2.0).vector(Point(4.0, 6.0))
        Vector(x=3.0, y=4.0)

        Args:
            other: The other point.

        Returns:
            The vector between the points.
        """
        other = self.aspoint(other)
        return Vector(other.x - self.x, other.y - self.y)

    @property
    def array(self) -> np.ndarray:
        """Convert the vector to a numpy array.

        >>> Point(1.0, 2.0).array
        array([1., 2.])

        Returns:
            The vector as a numpy array.
        """
        return np.array([self.x, self.y])

    @property
    def drawArgs(self) -> Tuple[float, float, float, float]:
        """Convert the point to a Qt point."""
        return self.x, self.y, 0.0, 0.0


class Vector(Point):
    def __matmul__(self, other: PointType) -> float:
        """Calculate the dot product between two points.

        >>> Vector(1.0, 2.0) @ Vector(3.0, 4.0)
        11.0

        Args:
            other: The other point.

        Returns:
            The dot product between the points.
        """
        other = self.aspoint(other)
        return self.x * other.x + self.y * other.y

    def dot(self, other: PointType) -> float:
        """Calculate the dot product between two points.

        >>> Vector(1.0, 2.0).dot(Vector(3.0, 4.0))
        11.0

        Args:
            other: The other point.

        Returns:
            The dot product between the points.
        """
        return self.__matmul__(other)

    @classmethod
    def asvector(cls, v: "Vector" | PointType) -> "Vector":
        """Convert a vector to a Vector object.

        >>> Vector.asvector(Vector(1.0, 2.0))
        Vector(x=1.0, y=2.0)
        >>> Vector.asvector((1.0, 2.0))
        Vector(x=1.0, y=2.0)
        >>> Vector.asvector(1.0 + 2.0j)
        Vector(x=1.0, y=2.0)

        Args:
            v: The vector to convert.

        Returns:
            The converted vector.
        """
        return cls.aspoint(v)

    @property
    def drawArgs(self) -> QLineF:
        """Convert the vector to a Qt line."""
        return QLineF(0.0, 0.0, self.x, self.y)



from __future__ import annotations

from typing import Iterable, List, Tuple

from qtpy.QtCore import QPointF
from qtpy.QtGui import QPolygonF
from shapely.geometry import Polygon as ShapelyPolygon

from .closedshape import ClosedShape
from .geometrybase import DrawAsPolygon
from .linebase import LineBase
from .point import Point, PointType
from .sequence import PointSequence
from .straightline import LineRay, LineSegment


class Polygon(ClosedShape, PointSequence):
    """A class representing a polygon."""

    drawAs = DrawAsPolygon

    def __init__(self, *points: PointType):
        """Create a polygon.

        >>> Polygon(Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0), Point(0.0, 1.0))
        Polygon(Point(x=0.0, y=0.0), Point(x=1.0, y=0.0), Point(x=1.0, y=1.0), Point(x=0.0, y=1.0))

        Args:
            points: The vertices of the polygon.
        """
        PointSequence.__init__(self, *points)

    def __repr__(self):
        """Return a string representation of the polygon.

        >>> Polygon(Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0), Point(0.0, 1.0))
        Polygon(Point(x=0.0, y=0.0), Point(x=1.0, y=0.0), Point(x=1.0, y=1.0), Point(x=0.0, y=1.0))
        """
        return PointSequence.__repr__(self)

    def __eq__(self, other: Polygon) -> bool:
        """Check if two polygons are equal.

        >>> polygon1 = Polygon(Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0), Point(0.0, 1.0))
        >>> polygon2 = Polygon(Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0), Point(0.0, 1.0))
        >>> polygon1 == polygon2
        True
        """
        return PointSequence.__eq__(self, other)

    @property
    def isSimple(self) -> bool:
        """Check if the polygon is simple.

        >>> Polygon(Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0), Point(0.0, 1.0)).isSimple
        True
        """
        return self.asShapely.is_simple

    @property
    def isValid(self):
        """Check if the polygon is convex."""
        return self.asShapely.is_valid

    def check(self):
        """Check if the polygon is valid."""
        assert len(self) >= 3, "A polygon must have at least three points."
        assert self.isSimple, "The polygon must be simple."

    @property
    def ndim(self) -> int:
        """Return the number of dimensions of the polygon.

        >>> Polygon(Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0)).ndim
        3
        """
        return len(self)

    @property
    def vertices(self) -> List[Point]:
        """Return the vertices of the polygon.

        >>> Polygon(Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0)).vertices
        [Point(x=0.0, y=0.0), Point(x=1.0, y=0.0), Point(x=1.0, y=1.0)]
        """
        return self.points

    @property
    def edges(self) -> List[LineSegment]:
        """Return the edges of the polygon.

        >>> edges = Polygon(Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0)).edges
        >>> edges[0]
        LineSegment(start=Point(x=0.0, y=0.0), end=Point(x=1.0, y=0.0)) -> LineSegment(A=0.0, B=1.0, C=0.0)
        >>> edges[1]
        LineSegment(start=Point(x=1.0, y=0.0), end=Point(x=1.0, y=1.0)) -> LineSegment(A=1.0, B=0.0, C=-1.0)
        >>> edges[2]
        LineSegment(start=Point(x=1.0, y=1.0), end=Point(x=0.0, y=0.0)) -> LineSegment(A=1.0, B=-1.0, C=0.0)
        """
        starts, ends = self.vertices, self.vertices[1:] + self.vertices[:1]
        return [LineSegment(start=start, end=end) for start, end in zip(starts, ends)]

    @property
    def boundaries(self) -> List[LineBase]:
        """Return the boundaries of the polygon.

        >>> lines = Polygon(Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0)).boundaries
        >>> lines[0]
        LineSegment(start=Point(x=0.0, y=0.0), end=Point(x=1.0, y=0.0)) -> LineSegment(A=0.0, B=1.0, C=0.0)
        >>> lines[1]
        LineSegment(start=Point(x=1.0, y=0.0), end=Point(x=1.0, y=1.0)) -> LineSegment(A=1.0, B=0.0, C=-1.0)
        >>> lines[2]
        LineSegment(start=Point(x=1.0, y=1.0), end=Point(x=0.0, y=0.0)) -> LineSegment(A=1.0, B=-1.0, C=0.0)
        """
        return self.edges

    def containsPoint(self, point: PointType | Iterable[Point]) -> bool:
        """Check if a point is inside the polygon.

        >>> polygon = Polygon(Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0), Point(0.0, 1.0))
        >>> (Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0), Point(0.0, 1.0)) in polygon
        True
        >>> Point(0.5, 0.5) in polygon
        True
        >>> (Point(0.5, 0.0), Point(0.5, 1.0), Point(0.0, 0.5), Point(1.0, 0.5)) in polygon
        True
        >>> Point(0.5, 1.5) in polygon
        False

        Args:
            point: A point or an iterable of points.
        """
        if isinstance(point, Iterable) and isinstance(tuple(point)[0], Point):
            return all(self.contains(p) for p in point)

        # Lie on the edges or vertices
        point = Point.aspoint(point)
        if self.inVertices(point) or self.inEdges(point):
            return True

        # LineRay casting
        ray = LineRay(start=point, end=Point(0, 0))
        intersections, intersecting_points = 0, PointSequence()
        for edge in self.edges:
            if ray.isIntersecting(edge):
                intersection = ray.intersect(edge)
                if intersection not in intersecting_points:
                    intersections += 1
                    intersecting_points.append(intersection)
        return intersections % 2 == 1

    def containsLine(self, line: LineBase | Iterable[LineBase]) -> bool:
        raise NotImplementedError

    def inEdges(self, item: PointType | Iterable[Point]) -> bool:
        """Check if a point is on the edges of the polygon.

        >>> polygon = Polygon(Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0), Point(0.0, 1.0))
        >>> polygon.inEdges((Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0), Point(0.0, 1.0)))
        True
        >>> polygon.inEdges((Point(0.5, 0.0), Point(0.5, 1.0), Point(0.0, 0.5), Point(1.0, 0.5)))
        True

        Args:
            item: A point or an iterable of points.
        """
        if isinstance(item, Iterable) and isinstance(tuple(item)[0], Point):
            return all(self.inEdges(p) for p in item)
        point = Point.aspoint(item)
        return any(point in edge for edge in self.edges)

    def inVertices(self, item: PointType | Iterable[Point]) -> bool:
        """Check if a point is on the vertices of the polygon.

        >>> polygon = Polygon(Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0), Point(0.0, 1.0))
        >>> polygon.inVertices((Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0), Point(0.0, 1.0)))
        True

        Args:
            item: A point or an iterable of points.
        """
        if isinstance(item, Iterable) and isinstance(tuple(item)[0], Point):
            return all(self.inVertices(p) for p in item)
        point = Point.aspoint(item)
        return any(point == vertex for vertex in self.vertices)

    @property
    def maxsize(self) -> float:
        """Return the maximum size of the polygon.

        >>> Polygon(Point(0.0, 0.0), Point(2.0, 0.0), Point(1.0, 1.0), Point(0.0, 1.0)).maxsize
        2.0
        """
        lf, tr = self.bbox
        return max(tr.x - lf.x, tr.y - lf.y)

    @property
    def area(self) -> float:
        """Return the area of the polygon.

        >>> Polygon(Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0), Point(0.0, 1.0)).area
        1.0
        """
        v1, v2 = self.vertices, self.vertices[1:] + self.vertices[:1]
        return abs(sum((p1.x - p2.x) * (p1.y + p2.y) for p1, p2 in zip(v1, v2))) / 2

    @property
    def perimeter(self) -> float:
        """Return the perimeter of the polygon.

        >>> assert Polygon(Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0), Point(0.0, 1.0)).perimeter == 4.0
        """
        return sum(edge.length for edge in self.edges)

    @property
    def centroid(self) -> Point:
        """Return the centroid of the polygon.

        >>> Polygon(Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0), Point(0.0, 1.0)).centroid
        Point(x=0.5, y=0.5)
        """
        v1, v2 = self.vertices, self.vertices[1:] + self.vertices[:1]
        x = abs(sum((p1.x + p2.x) * (p1.x * p2.y - p2.x * p1.y) for p1, p2 in zip(v1, v2))) / (6 * self.area)
        y = abs(sum((p1.y + p2.y) * (p1.x * p2.y - p2.x * p1.y) for p1, p2 in zip(v1, v2))) / (6 * self.area)
        return Point(x, y)

    @property
    def bbox(self) -> Tuple[Point, Point]:
        """Return the bounding box of the polygon.

        >>> Polygon(Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0), Point(0.0, 1.0)).bbox
        (Point(x=0.0, y=0.0), Point(x=1.0, y=1.0))
        """
        x, y = zip(*self.vertices)
        return Point(min(x), min(y)), Point(max(x), max(y))

    @property
    def asShapely(self) -> ShapelyPolygon:
        """Return the polygon as a Shapely polygon.

        >>> Polygon(Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0), Point(0.0, 1.0)).asShapely
        <POLYGON ...>
        """
        return ShapelyPolygon(((vertex.x, vertex.y) for vertex in self.vertices))

    @property
    def drawArgs(self) -> Tuple[QPolygonF]:
        """Return the polygon as a Qt polygon."""
        return (QPolygonF([QPointF(vertex.x, vertex.y) for vertex in self.vertices]),)



from __future__ import annotations

from svgpathtools.path import Path

from .filterbase import FilterBase


class ClosedPathFilter(FilterBase):
    """Filter for closed paths."""

    def accept(self, path: Path) -> bool:
        return not self.enabled or not (path.iscontinuous() and path.isclosed())



from .arc import Arc
from .bezier import Bezier, CubicBezier, QuadraticBezier
from .circle import Circle
from .closedshape import ClosedShape
from .ellipse import Ellipse
from .geometrybase import GeometryBase
from .linebase import CurveLineBase, LineBase, StraightLineBase
from .path import Path, PathSequence
from .point import Point, PointType, Vector
from .polygon import Polygon
from .quadrilateral import (
    IsoscelesTrapezoid,
    Kite,
    Parallelogram,
    Quadrilateral,
    Rectangle,
    RightKite,
    Square,
    Trapezoid,
)
from .straightline import (
    ExtendedLineRay,
    ExtendedLineSegment,
    Line,
    LineRay,
    LineSegment,
)
from .triangle import Triangle

__all__ = [
    "Arc",
    "Bezier",
    "CubicBezier",
    "QuadraticBezier",
    "Circle",
    "ClosedShape",
    "Ellipse",
    "GeometryBase",
    "LineBase",
    "StraightLineBase",
    "CurveLineBase",
    "Path",
    "PathSequence",
    "Point",
    "Vector",
    "PointType",
    "Polygon",
    "Quadrilateral",
    "Trapezoid",
    "Parallelogram",
    "Rectangle",
    "Square",
    "Kite",
    "RightKite",
    "IsoscelesTrapezoid",
    "Line",
    "LineSegment",
    "LineRay",
    "ExtendedLineSegment",
    "ExtendedLineRay",
    "Triangle",
]



from abc import ABC
from typing import List

from svgpathtools import CubicBezier as SvgPathToolsCubicBezier
from svgpathtools import QuadraticBezier as SvgPathToolsQuadraticBezier

from .linebase import CurveLineBase
from .point import Point, PointType


class Bezier(CurveLineBase, ABC):
    """A class representing a Bézier curve."""

    #: The start point of the quadratic Bézier curve.
    start: Point
    #: The control point of the quadratic Bézier curve.
    controls: List[Point]
    #: The end point of the quadratic Bézier curve.
    end: Point

    def __init__(self, start: PointType, controls: List[PointType], end: PointType):
        self.start = Point.aspoint(start)
        self.controls = [Point.aspoint(control) for control in controls]
        self.end = Point.aspoint(end)

    def __repr__(self):
        return f"{self.__class__.__name__}(start={self.start}, controls={self.controls}, end={self.end})"

    def __iter__(self):
        yield self.start
        for control in self.controls:
            yield control
        yield self.end


class QuadraticBezier(Bezier):
    def __init__(self, start: PointType, control: PointType, end: PointType):
        super().__init__(start, [control], end)

    @classmethod
    def fromSvgPathToolsQuadraticBezier(cls, bezier: SvgPathToolsQuadraticBezier):
        return cls(Point(bezier.start), Point(bezier.control), Point(bezier.end))


class CubicBezier(Bezier):
    def __init__(self, start: PointType, control1: PointType, control2: PointType, end: PointType):
        super().__init__(start, [control1, control2], end)

    @classmethod
    def fromSvgPathToolsCubicBezier(cls, bezier: SvgPathToolsCubicBezier):
        return cls(Point(bezier.start), Point(bezier.control1), Point(bezier.control2), Point(bezier.end))



from __future__ import annotations

from abc import ABC
from typing import Iterable, List, Tuple

from .geometrybase import GeometryBase
from .linebase import LineBase
from .point import Point


class ClosedShape(GeometryBase, ABC):
    """Base class for closed shapes."""

    def __eq__(self, other: "ClosedShape") -> bool:
        raise NotImplementedError

    def __contains__(self, item: GeometryBase | Iterable["GeometryBase"]) -> bool:
        """Check if a point or shape is inside the shape."""
        return self.contains(item)

    def contains(self, item: GeometryBase | Iterable["GeometryBase"]) -> bool:
        """Check if a point or shape is inside the shape."""
        if isinstance(item, Point):
            return self.containsPoint(item)
        elif isinstance(item, LineBase):
            return self.containsLine(item)
        elif isinstance(item, ClosedShape):
            return all(self.containsLine(line) for line in item.boundaries)
        elif isinstance(item, Iterable):
            return all(self.contains(subitem) for subitem in item)
        else:
            raise TypeError(f"Unsupported type: {type(item)}")

    @property
    def boundaries(self) -> List[LineBase]:
        """Return the boundaries of the shape. To decide whether the shape is inside another shape."""
        raise NotImplementedError

    def containsPoint(self, point: Point | Iterable[Point]) -> bool:
        """Check if a point is inside the shape."""
        raise NotImplementedError

    def containsLine(self, line: LineBase | Iterable[LineBase]) -> bool:
        """Check if a line is inside the shape."""
        raise NotImplementedError

    @property
    def area(self) -> float:
        """Return the area of the shape."""
        raise NotImplementedError

    @property
    def perimeter(self) -> float:
        """Return the perimeter of the shape."""
        raise NotImplementedError

    @property
    def centroid(self) -> Point:
        """Return the centroid of the shape."""
        raise NotImplementedError

    @property
    def bbox(self) -> Tuple[Point, Point]:
        """Return the bounding box of the shape."""
        raise NotImplementedError



from __future__ import annotations

import sys
from abc import ABC
from enum import IntEnum
from typing import Tuple

from qtpy.QtCore import Qt
from qtpy.QtGui import QPen
from qtpy.QtWidgets import QApplication, QGraphicsScene

from ..graphics.annotations import QBrushType, QGraphicsItemType, QPenType


class GeometryDrawAs(IntEnum):
    """Geometry draw as enum."""

    DrawAsLine = 0
    DrawAsPolygon = 1
    DrawAsEllipse = 2


DrawAsLine = GeometryDrawAs.DrawAsLine
DrawAsPolygon = GeometryDrawAs.DrawAsPolygon
DrawAsEllipse = GeometryDrawAs.DrawAsEllipse


class GeometryBase(ABC):
    """A base class for all geometries."""

    #: Tolerance for equality.
    tolerance = 1e-6

    #: type of drawing
    drawAs: GeometryDrawAs

    def __repr__(self) -> str:
        """Return the representation of the geometry."""
        raise NotImplementedError

    def __eq__(self, other: "GeometryBase") -> bool:
        """Check if two geometries are equal."""
        raise NotImplementedError

    def __ne__(self, other: "GeometryBase") -> bool:
        """Check if two geometries are not equal."""
        return not self.__eq__(other)

    @property
    def maxsize(self) -> float:
        """Return the max size of the geometry."""
        raise NotImplementedError

    @property
    def drawArgs(self) -> Tuple[str, tuple]:
        """Return the arguments for drawing the geometry."""
        raise NotImplementedError

    def draw(
        self,
        scene: QGraphicsScene,
        pen: QPenType = None,
        brush: QBrushType = None,
        item: QGraphicsItemType = None,
    ) -> QGraphicsItemType:
        """Draw the geometry on the scene.

        Args:
            scene: The scene to draw on.
            pen: The pen to draw with.
            brush: The brush to draw with.
            item: The old item to draw on, if any.
        """
        args = self.drawArgs
        if self.drawAs == DrawAsLine:
            item and item.setLine(*args) or item or (item := scene.addLine(*args))
        elif self.drawAs == DrawAsPolygon:
            item and item.setPolygon(*args) or item or (item := scene.addPolygon(*args))
        elif self.drawAs == DrawAsEllipse:
            item and item.setRect(*args) or item or (item := scene.addEllipse(*args))
        else:
            raise ValueError(f"Unknown type {self.drawAs}")
        pen = QPen(pen or item.pen())
        pen.setWidthF(min(0.02 * self.maxsize, pen.widthF()))
        item.setPen(pen)
        item.setBrush(brush or item.brush())
        return item

    def plot(self, pen: QPenType = None, brush: QBrushType = None, fit: bool = True):
        """Plot the geometry.

        Args:
            pen: The pen to draw with.
            brush: The brush to draw with.
            fit: Fit the view to the geometry.
        """
        from ..graphics.graphicsview import GraphicsView

        app = QApplication(sys.argv)
        scene = QGraphicsScene()
        view = GraphicsView(scene)
        view.setWindowTitle(repr(self))
        view.resize(800, 600)
        view.show()
        self.draw(scene, pen, brush)
        fit and view.fitInView(scene.itemsBoundingRect(), Qt.KeepAspectRatio)
        sys.exit(app.exec_())
